from __future__ import annotations

from typing import (
    Any,
    Iterable,
    Literal,
    Optional,
    Self,
    Sequence,
    TypeVar,
    Union,
    overload,
)
from weakref import WeakKeyDictionary

from sqlalchemy import exc, inspect, tuple_, util
from sqlalchemy.engine.cursor import CursorResult
from sqlalchemy.engine.interfaces import _CoreAnyExecuteParams, _CoreSingleExecuteParams
from sqlalchemy.engine.result import Result, ScalarResult
from sqlalchemy.ext.asyncio import AsyncResult
from sqlalchemy.ext.asyncio import AsyncSession as SqlalchemyAsyncSession
from sqlalchemy.orm import (
    InstanceState,
    Query,
    attributes,
    object_mapper,
)
from sqlalchemy.orm import Session as SqlalchemySession
from sqlalchemy.orm._typing import (
    OrmExecuteOptionsParameter,
    _IdentityKeyType,
)
from sqlalchemy.orm.base import _O, _state_mapper
from sqlalchemy.orm.interfaces import ORMOption
from sqlalchemy.orm.session import (
    JoinTransactionMode,
    _BindArguments,
    _EntityBindKey,
    _PKIdentityArgument,
    _SessionBind,
    _SessionBindKey,
)
from sqlalchemy.orm.util import Bundle
from sqlalchemy.sql import Executable, Select, functions, select
from sqlalchemy.sql._typing import (
    _ColumnExpressionArgument,
    _ColumnExpressionOrStrLabelArgument,
    _InfoType,
)
from sqlalchemy.sql.base import ExecutableOption, _NoArg
from sqlalchemy.sql.dml import Delete, Insert, Update, UpdateBase
from sqlalchemy.sql.selectable import ForUpdateArg, ForUpdateParameter, TypedReturnsRows

from arcanus.base import (
    BaseTransmuter,
    Transmuter,
    TransmuterProxied,
    ValidateContextGeneratorT,
    ValidationContextT,
    validation_context,
)
from arcanus.materia.base import active_materia
from arcanus.materia.sqlalchemy.result import (
    _T,
    AdaptedResult,
    AsyncAdaptedResult,
)

T = TypeVar("T", bound=Transmuter)


def resolve_statement_entities(statement: Executable) -> list[type[Any]]:
    entities: list[type[Any]] = []
    if isinstance(statement, Select):
        for desc in statement.column_descriptions:
            if ((expr := desc.get("expr")) is not None) and (
                (type := desc.get("type")) is not None
            ):
                # Bundle types (For example, used by selectinload for pk grouping) return tuple[*]
                if type is Bundle:
                    entities.append(tuple[*(e.type.python_type for e in expr.exprs)])
                else:
                    transmuter = active_materia.get().formulars.reverse.get(type)
                    if transmuter:
                        entities.append(transmuter)
                    else:
                        try:
                            entities.append(type.python_type)
                        except NotImplementedError:
                            # NullType and other types without python_type
                            entities.append(object)
    elif isinstance(statement, (Insert, Update, Delete)):
        if statement._returning:
            for item in statement._returning:
                if transmuter := active_materia.get().formulars.reverse.get(
                    item.entity_namespace
                ):
                    entities.append(transmuter)
                else:
                    try:
                        entities.append(item.type.python_type)  # type: ignore[attr-defined]
                    except NotImplementedError:
                        # NullType and other types without python_type
                        entities.append(object)
    return entities


class Session(SqlalchemySession):
    _validation_context: ValidationContextT
    _validation_context_manager: ValidateContextGeneratorT | None

    def __init__(
        self,
        bind: Optional[_SessionBind] = None,
        *,
        autoflush: bool = True,
        future: Literal[True] = True,
        expire_on_commit: bool = True,
        autobegin: bool = True,
        twophase: bool = False,
        binds: Optional[dict[_SessionBindKey, _SessionBind]] = None,
        enable_baked_queries: bool = True,
        info: Optional[_InfoType] = None,
        query_cls: Optional[type[Query[Any]]] = None,
        autocommit: Literal[False] = False,
        join_transaction_mode: JoinTransactionMode = "conditional_savepoint",
        close_resets_only: Union[bool, _NoArg] = _NoArg.NO_ARG,
    ) -> None:
        super().__init__(
            bind,
            autoflush=autoflush,
            future=future,
            expire_on_commit=expire_on_commit,
            autobegin=autobegin,
            twophase=twophase,
            binds=binds,
            enable_baked_queries=enable_baked_queries,
            info=info,
            query_cls=query_cls,
            autocommit=autocommit,
            join_transaction_mode=join_transaction_mode,
            close_resets_only=close_resets_only,
        )
        self._validation_context = WeakKeyDictionary()
        self._validation_context_manager = None

    def __enter__(self):
        self._validation_context_manager = validation_context(self._validation_context)
        self._validation_context_manager.__enter__()
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> Optional[bool]:
        if self._validation_context_manager is not None:
            self._validation_context_manager.__exit__(exc_type, exc_value, traceback)
        return super().__exit__(exc_type, exc_value, traceback)

    def __iter__(self) -> Iterable[BaseTransmuter]:
        if not self._validation_context:
            raise RuntimeError(
                "Active validation context is requried, please use a context manager 'with Session() as session' to create a session context."
            )
        for instance in super().__iter__():
            if instance in self._validation_context:
                yield self._validation_context[instance]
            if isinstance(instance, TransmuterProxied) and (
                transmuter := instance.transmuter_proxy
            ):
                yield transmuter
            yield instance  # type: ignore[reportUnreachable]

    @overload
    def execute(
        self,
        statement: TypedReturnsRows[_T],
        params: Optional[_CoreAnyExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        _parent_execute_state: Optional[Any] = None,
        _add_event: Optional[Any] = None,
    ) -> Result[_T]: ...

    @overload
    def execute(
        self,
        statement: UpdateBase,
        params: Optional[_CoreAnyExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        _parent_execute_state: Optional[Any] = None,
        _add_event: Optional[Any] = None,
    ) -> CursorResult[Any]: ...
    @overload
    def execute(
        self,
        statement: Executable,
        params: Optional[_CoreAnyExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        _parent_execute_state: Optional[Any] = None,
        _add_event: Optional[Any] = None,
    ) -> Result[Any]: ...
    def execute(
        self,
        statement: Executable,
        params: Optional[_CoreAnyExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        _parent_execute_state: Optional[Any] = None,
        _add_event: Optional[Any] = None,
    ) -> Result[Any]:
        result = super().execute(
            statement,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            _parent_execute_state=_parent_execute_state,
            _add_event=_add_event,
        )

        if execution_options.get("sa_top_level_orm_context", False):
            return result

        if execution_options.get("_sa_orm_load_options", {}):
            return result

        entities = resolve_statement_entities(statement)
        if entities and any(
            isinstance(e, type) and issubclass(e, Transmuter) for e in entities
        ):
            return AdaptedResult(
                real_result=result,
                entities=tuple(entities),
            )  # pyright: ignore[reportReturnType]

        return result

    @overload
    def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        params: Optional[_CoreSingleExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        **kw: Any,
    ) -> Optional[_T]: ...
    @overload
    def scalar(
        self,
        statement: Executable,
        params: Optional[_CoreSingleExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        **kw: Any,
    ) -> Any: ...
    def scalar(
        self,
        statement: Executable,
        params: Optional[_CoreSingleExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        **kw: Any,
    ) -> Any:
        return self.execute(
            statement=statement,
            params=params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            **kw,
        ).scalar()

    @overload
    def scalars(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        params: Optional[_CoreAnyExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        **kw: Any,
    ) -> ScalarResult[_T]: ...
    @overload
    def scalars(
        self,
        statement: Executable,
        params: Optional[_CoreAnyExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        **kw: Any,
    ) -> ScalarResult[Any]: ...
    def scalars(
        self,
        statement: Executable,
        params: Optional[_CoreAnyExecuteParams] = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
        **kw: Any,
    ) -> ScalarResult[Any]:
        return self.execute(
            statement=statement,
            params=params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            **kw,
        ).scalars()

    def expunge(self, instance: object) -> None:
        if isinstance(instance, Transmuter):
            if instance.__transmuter_provided__ in self._validation_context:
                del self._validation_context[instance.__transmuter_provided__]
            super().expunge(instance.__transmuter_provided__)
        else:
            super().expunge(instance)

    def expunge_all(self) -> None:
        self._validation_context.clear()
        return super().expunge_all()

    def add(self, instance: object, _warn: bool = True) -> None:
        if isinstance(instance, Transmuter):
            self._validation_context[instance.__transmuter_provided__] = instance
            super().add(instance.__transmuter_provided__, _warn=_warn)
        else:
            super().add(instance, _warn=_warn)

    def _save_or_update_state(
        self,
        state: InstanceState,
    ) -> None:
        state._orphaned_outside_of_session = False
        self._save_or_update_impl(state)

        mapper = _state_mapper(state)
        for o, m, st_, dct_ in mapper.cascade_iterator(
            "save-update", state, halt_on=self._contains_state
        ):
            if isinstance(o, TransmuterProxied) and (transmuter := o.transmuter_proxy):
                self._validation_context[o] = transmuter
            self._save_or_update_impl(st_)

    def refresh(
        self,
        instance: object,
        attribute_names: Iterable[str] | None = None,
        with_for_update: ForUpdateArg | None | bool | dict[str, Any] = None,
    ) -> None:
        if isinstance(instance, Transmuter):
            if instance.__transmuter_provided__ not in self._validation_context:
                self._validation_context[instance.__transmuter_provided__] = instance
            super().refresh(instance, attribute_names, with_for_update)
            instance.revalidate()
        else:
            super().refresh(instance, attribute_names, with_for_update)

    def rollback(self) -> None:
        super().rollback()
        self._validation_context.clear()

    def merge(
        self,
        instance: T,
        *,
        load: bool = True,
        options: Sequence[ORMOption] | None = None,
    ) -> T:
        if isinstance(instance, Transmuter):
            if self._warn_on_events:
                self._flush_warning("Session.merge()")

            _recursive: dict[InstanceState[Any], object] = {}
            _resolve_conflict_map: dict[_IdentityKeyType[Any], object] = {}

            if load:
                # flush current contents if we expect to load data
                self._autoflush()

            object_mapper(instance)  # verify mapped
            autoflush = self.autoflush
            try:
                self.autoflush = False
                merged = self._merge(
                    attributes.instance_state(instance),
                    attributes.instance_dict(instance.__transmuter_provided__),
                    load=load,
                    options=options,
                    _recursive=_recursive,
                    _resolve_conflict_map=_resolve_conflict_map,
                )
                instance = type(instance).model_validate(merged)
                instance.revalidate()
                return instance
            finally:
                self.autoflush = autoflush
        else:
            return super().merge(
                instance,
                load=load,
                options=options,
            )

    def enable_relationship_loading(self, obj: BaseTransmuter) -> None:
        super().enable_relationship_loading(obj.__transmuter_provided__)
        self._validation_context[obj.__transmuter_provided__] = obj

    @overload
    def get(
        self,
        entity: type[T],
        ident: _PKIdentityArgument,
        *,
        options: Sequence[ORMOption] | None = None,
        populate_existing: bool = False,
        with_for_update: ForUpdateParameter = None,
        identity_token: Optional[Any] = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
    ) -> Optional[T]: ...
    @overload
    def get(
        self,
        entity: _EntityBindKey[_O],
        ident: _PKIdentityArgument,
        *,
        options: Sequence[ORMOption] | None = None,
        populate_existing: bool = False,
        with_for_update: ForUpdateParameter = None,
        identity_token: Optional[Any] = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
    ) -> Optional[_O]: ...
    def get(
        self,
        entity: type[T] | _EntityBindKey[_O],
        ident: _PKIdentityArgument,
        *,
        options: Sequence[ORMOption] | None = None,
        populate_existing: bool = False,
        with_for_update: ForUpdateParameter = None,
        identity_token: Optional[Any] = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
    ) -> Optional[T] | Optional[_O]:
        if isinstance(entity, type) and issubclass(entity, Transmuter):
            instance = super().get(
                # sqlalchemy materia requires transumter to have a provider blessed
                entity.__transmuter_provider__,  # pyright: ignore[reportArgumentType]
                ident,
                options=options,
                populate_existing=populate_existing,
                with_for_update=with_for_update,
                identity_token=identity_token,
                execution_options=execution_options,
                bind_arguments=bind_arguments,
            )
            if not instance:
                return None
            return entity.model_validate(instance)
        else:
            instance = super().get(
                entity,
                ident,
                options=options,
                populate_existing=populate_existing,
                with_for_update=with_for_update,
                identity_token=identity_token,
                execution_options=execution_options,
                bind_arguments=bind_arguments,
            )
            if isinstance(instance, Transmuter):
                return instance.__transmuter_provided__  # pyright: ignore[reportReturnType]
            return instance

    @overload
    def get_one(
        self,
        entity: type[T],
        ident: _PKIdentityArgument,
        *,
        options: Optional[Sequence[ORMOption]] = None,
        populate_existing: bool = False,
        with_for_update: ForUpdateParameter = None,
        identity_token: Optional[Any] = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
    ) -> T: ...
    @overload
    def get_one(
        self,
        entity: _EntityBindKey[_O],
        ident: _PKIdentityArgument,
        *,
        options: Optional[Sequence[ORMOption]] = None,
        populate_existing: bool = False,
        with_for_update: ForUpdateParameter = None,
        identity_token: Optional[Any] = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
    ) -> _O: ...
    def get_one(
        self,
        entity: type[T] | _EntityBindKey[_O],
        ident: _PKIdentityArgument,
        *,
        options: Optional[Sequence[ORMOption]] = None,
        populate_existing: bool = False,
        with_for_update: ForUpdateParameter = None,
        identity_token: Optional[Any] = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: Optional[_BindArguments] = None,
    ) -> T | _O:
        instance = self.get(
            entity,
            ident,
            options=options,
            populate_existing=populate_existing,
            with_for_update=with_for_update,
            identity_token=identity_token,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
        )

        if instance is None:
            raise exc.NoResultFound("No row was found when one was required")

        return instance

    def one(
        self,
        entity: type[_T],
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ):
        statement = select(entity)

        if expressions:
            statement = statement.where(*expressions)
        if filters:
            statement = statement.filter_by(**filters)
        if options:
            statement = statement.options(*options)
        if execution_options:
            statement = statement.execution_options(**execution_options)

        return self.execute(statement).scalar_one()

    def one_or_none(
        self,
        entity: type[_T],
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ):
        statement = select(entity)

        if expressions:
            statement = statement.where(*expressions)
        if filters:
            statement = statement.filter_by(**filters)
        if options:
            statement = statement.options(*options)
        if execution_options:
            statement = statement.execution_options(**execution_options)

        return self.execute(statement).scalar_one_or_none()

    def first(
        self,
        entity: type[_T],
        order_bys: Iterable[_ColumnExpressionOrStrLabelArgument[Any]] | None = None,
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ):
        statement = select(entity)

        if order_bys:
            statement = statement.order_by(*order_bys)
        if expressions:
            statement = statement.where(*expressions)
        if filters:
            statement = statement.filter_by(**filters)
        if options:
            statement = statement.options(*options)
        if execution_options:
            statement = statement.execution_options(**execution_options)

        return self.execute(statement).scalars().first()

    def bulk(
        self,
        entity: type[_T],
        idents: Sequence[_PKIdentityArgument],
        *,
        options: Sequence[ORMOption] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
    ) -> list[_T | None]:
        """Bulk version of Session.get. Each element in idents should be
        exactly the same format as Session.get's ident parameter.

        Returns a list of entities in the same order as idents, with None
        for any ident that was not found.
        """
        if not idents:
            return []

        mapper = inspect(entity.__transmuter_provider__)
        pk_columns = mapper.primary_key  # pyright: ignore[reportOptionalMemberAccess]

        if len(pk_columns) == 1:
            # Build the WHERE clause based on single or composite PK
            pk_col = pk_columns[0]
            statement = select(entity).where(pk_col.in_(idents))
        else:
            # Composite PK: use tuple comparison
            # Each ident should be a tuple matching the PK columns
            statement = select(entity).where(tuple_(*pk_columns).in_(idents))

        if options:
            statement = statement.options(*options)
        if execution_options:
            statement = statement.execution_options(**execution_options)

        entities = self.execute(statement).scalars().all()

        # Build mapping from PK value(s) to entity
        if len(pk_columns) == 1:
            pk_attr = pk_columns[0].key
            mapping = {getattr(e, pk_attr): e for e in entities}
            return [mapping.get(ident) for ident in idents]
        else:
            # Composite PK: map tuple of PK values to entity
            pk_attrs = [col.key for col in pk_columns]
            mapping = {
                tuple(getattr(e, attr) for attr in pk_attrs): e for e in entities
            }
            return [
                mapping.get(tuple(ident) if not isinstance(ident, tuple) else ident)
                for ident in idents
            ]

    def count(
        self,
        entity: type[_T],
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ):
        statement = select(functions.count()).select_from(entity)

        if expressions:
            statement = statement.where(*expressions)
        if filters:
            statement = statement.filter_by(**filters)
        if execution_options:
            statement = statement.execution_options(**execution_options)

        return self.execute(statement).scalar_one()

    def list(
        self,
        entity: type[_T],
        limit: int | None = 100,
        offset: int | None = None,
        # cursor: UUID | None = None, # TODO: re-enable cursor pagination when identity solution is clarified
        order_bys: Iterable[_ColumnExpressionOrStrLabelArgument[Any]] | None = None,
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ):
        statement = select(entity)

        if limit:
            statement = statement.limit(limit)
        if offset:
            statement = statement.offset(offset)
        if order_bys:
            statement = statement.order_by(*order_bys)
        if options:
            statement = statement.options(*options)
        if expressions:
            statement = statement.where(*expressions)
        if execution_options:
            statement = statement.execution_options(**execution_options)
        if filters:
            statement = statement.filter_by(**filters)

        return self.execute(statement).scalars().all()

    def partitions(
        self,
        entity: type[_T],
        limit: int | None = 100,
        offset: int | None = None,
        # cursor: UUID | None = None, # TODO: re-enable cursor pagination when identity solution is clarified
        size: int | None = 10,
        order_bys: Iterable[_ColumnExpressionOrStrLabelArgument[Any]] | None = None,
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ):
        statement = select(entity).execution_options(yield_per=size)

        if limit:
            statement = statement.limit(limit)
        if offset:
            statement = statement.offset(offset)
        if order_bys:
            statement = statement.order_by(*order_bys)
        if options:
            statement = statement.options(*options)
        if expressions:
            statement = statement.where(*expressions)
        if execution_options:
            statement = statement.execution_options(**execution_options)
        if filters:
            statement = statement.filter_by(**filters)

        yield from self.execute(statement).scalars().partitions(size)


class AsyncSession(SqlalchemyAsyncSession):
    sync_session_class = Session
    sync_session: Session

    async def __aenter__(self) -> Self:
        self._validation_context_manager = validation_context(self._validation_context)
        self._validation_context_manager.__enter__()
        await super().__aenter__()
        return self

    async def __aexit__(self, type_: Any, value: Any, traceback: Any) -> None:
        if self._validation_context_manager is not None:
            self._validation_context_manager.__exit__(type_, value, traceback)
            self._validation_context_manager = None
        await super().__aexit__(type_, value, traceback)

    @property
    def _validation_context(self) -> ValidationContextT:
        return self.sync_session._validation_context

    @property
    def _validation_context_manager(self) -> ValidateContextGeneratorT | None:
        return self.sync_session._validation_context_manager

    @_validation_context_manager.setter
    def _validation_context_manager(self, value: ValidateContextGeneratorT | None):
        self.sync_session._validation_context_manager = value

    @overload
    async def stream(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> AsyncResult[tuple[_T]]: ...

    @overload
    async def stream(
        self,
        statement: Executable,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> AsyncResult[Any]: ...

    async def stream(
        self,
        statement: Executable,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> AsyncResult[Any]:
        """Execute a statement and return a streaming
        :class:`AsyncAdaptedResult` object that adapts rows to transmuter types.
        """
        _STREAM_OPTIONS = util.immutabledict({"stream_results": True})

        if execution_options:
            execution_options = util.immutabledict(execution_options).union(
                _STREAM_OPTIONS
            )
        else:
            execution_options = _STREAM_OPTIONS

        result = await util.greenlet_spawn(
            self.sync_session.execute,
            statement,
            params=params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            **kw,
        )

        if isinstance(result, AdaptedResult):
            return AsyncAdaptedResult(
                result,
                entities=result.entities,
            )  # pyright: ignore[reportReturnType]

        return AsyncResult(result)

    async def one(
        self,
        entity: type[_T],
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ) -> _T:
        return await util.greenlet_spawn(
            self.sync_session.one,
            entity,
            options,
            expressions,
            execution_options,
            **filters,
        )

    async def one_or_none(
        self,
        entity: type[_T],
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ) -> Optional[_T]:
        r = await util.greenlet_spawn(
            self.sync_session.one_or_none,
            entity,
            options,
            expressions,
            execution_options,
            **filters,
        )
        return r

    async def first(
        self,
        entity: type[_T],
        order_bys: Iterable[_ColumnExpressionOrStrLabelArgument[Any]] | None = None,
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ) -> Optional[_T]:
        return await util.greenlet_spawn(
            self.sync_session.first,
            entity,
            order_bys,
            options,
            expressions,
            execution_options,
            **filters,
        )

    async def bulk(
        self,
        entity: type[_T],
        idents: Sequence[_PKIdentityArgument],
        *,
        options: Sequence[ORMOption] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
    ) -> list[_T | None]:
        return await util.greenlet_spawn(
            self.sync_session.bulk,
            entity,
            idents,
            options=options,
            execution_options=execution_options,
        )

    async def count(
        self,
        entity: type[_T],
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ) -> int:
        return await util.greenlet_spawn(
            self.sync_session.count,
            entity,
            expressions,
            execution_options,
            **filters,
        )

    async def list(
        self,
        entity: type[_T],
        limit: int | None = 100,
        offset: int | None = None,
        order_bys: Iterable[_ColumnExpressionOrStrLabelArgument[Any]] | None = None,
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ) -> Sequence[_T]:
        return await util.greenlet_spawn(
            self.sync_session.list,
            entity,
            limit,
            offset,
            order_bys,
            options,
            expressions,
            execution_options,
            **filters,
        )

    async def partitions(
        self,
        entity: type[_T],
        limit: int | None = 100,
        offset: int | None = None,
        size: int | None = 10,
        order_bys: Iterable[_ColumnExpressionOrStrLabelArgument[Any]] | None = None,
        options: Iterable[ExecutableOption] | None = None,
        expressions: Iterable[_ColumnExpressionArgument[bool]] | None = None,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        **filters,
    ):
        statement = select(entity)

        if limit:
            statement = statement.limit(limit)
        if offset:
            statement = statement.offset(offset)
        if order_bys:
            statement = statement.order_by(*order_bys)
        if options:
            statement = statement.options(*options)
        if expressions:
            statement = statement.where(*expressions)
        if execution_options:
            statement = statement.execution_options(**execution_options)
        if filters:
            statement = statement.filter_by(**filters)

        async for partition in (
            (
                await self.stream(
                    statement,
                    execution_options=dict(yield_per=size),
                )
            )
            .scalars()
            .partitions(size)
        ):
            yield partition



"""Test custom Session features.

Tests:
- Session context management
- get_one behavior (raises if not found)
- Object identity and caching
- Session.add and Session.add_all
- Session.delete
- Session.flush and Session.commit
- Session.rollback
- Session.expire and Session.refresh
- Session.expunge
"""

from __future__ import annotations

import uuid

import pytest
from sqlalchemy import Engine, select, update
from sqlalchemy.exc import NoResultFound

from arcanus.base import BaseTransmuter, Transmuter
from arcanus.materia.sqlalchemy import Session
from tests.transmuters import Author, Book, BookCategory, Category, Publisher


class TestSessionContextManagement:
    """Test Session context management."""

    def test_session_context_manager(self, engine: Engine):
        """Test using Session as a context manager."""
        with Session(engine) as session:
            author = Author(name="Context Test", field="Physics")
            session.add(author)
            session.flush()
            author.revalidate()

            assert author.id is not None

        # Session should be closed outside context

    def test_session_commit_on_exit(self, engine: Engine):
        """Test that changes are committed on successful context exit."""
        author_id = None
        with Session(engine) as session:
            with session.begin():
                author = Author(name="Auto Commit Test", field="Biology")
                session.add(author)
                session.flush()
                author.revalidate()
                author_id = author.id

        # Verify in new session
        with Session(engine) as session:
            retrieved = session.get_one(Author, author_id)
            assert retrieved.name == "Auto Commit Test"

    def test_session_rollback_on_exception(self, engine: Engine):
        """Test that changes are rolled back on exception."""
        author_id = 999999
        try:
            with Session(engine) as session:
                author = Author(id=author_id, name="Rollback Test", field="Chemistry")
                session.add(author)
                session.flush()
                author_id = author.id

                # Cause an exception
                raise ValueError("Test exception")
        except ValueError:
            pass

        # Changes should be rolled back
        with Session(engine) as session:
            result = session.get(Author, author_id)
            assert result is None

    def test_nested_context_managers(self, engine: Engine):
        """Test nested Session context managers (not recommended but possible)."""
        with Session(engine) as outer_session:
            author = Author(name="Outer", field="Physics")
            outer_session.add(author)
            outer_session.flush()
            author.revalidate()
            outer_session.commit()

            author_id = author.id

            with Session(engine) as inner_session:
                # Inner session can see committed data
                inner_author = inner_session.get_one(Author, author_id)
                assert inner_author.name == "Outer"


class TestSessionHelpers:
    """Test Session helper methods: get, get_one, one, one_or_none, first, bulk, count, list, partitions."""

    def test_get_one_success(self, engine: Engine):
        """Test get_one returns object when found."""
        with Session(engine) as session:
            author = Author(name="Get One Test", field="Physics")
            session.add(author)
            session.flush()
            author.revalidate()
            author_id = author.id

            # Retrieve
            retrieved = session.get_one(Author, author_id)
            assert retrieved.id == author_id
            assert retrieved.name == "Get One Test"

    def test_get_one_raises_when_not_found(self, engine: Engine):
        """Test get_one raises NoResultFound when not found."""
        with Session(engine) as session:
            with pytest.raises(NoResultFound):
                session.get_one(Author, 999999)

    def test_get_one_vs_get(self, engine: Engine):
        """Test difference between get_one and get."""
        with Session(engine) as session:
            # get returns None
            result = session.get(Author, 999999)
            assert result is None

            # get_one raises
            with pytest.raises(NoResultFound):
                session.get_one(Author, 999999)

    def test_one_success(self, engine: Engine):
        """Test one returns single matching entity."""
        with Session(engine) as session:
            author1 = Author(name="One Test 1", field="Physics")
            author2 = Author(name="One Test 2", field="Biology")
            session.add_all([author1, author2])
            session.flush()
            author1.revalidate()

            # Query by filter
            result = session.one(Author, name="One Test 1")
            assert result.id == author1.id
            assert result.name == "One Test 1"

    def test_one_with_expressions(self, engine: Engine):
        """Test one with where expressions."""
        with Session(engine) as session:
            author1 = Author(name="Expression Test 1", field="Physics")
            author2 = Author(name="Expression Test 2", field="Physics")
            session.add_all([author1, author2])
            session.flush()
            author1.revalidate()

            # Query with expression
            result = session.one(
                Author, expressions=[Author["name"] == "Expression Test 1"]
            )
            assert result.id == author1.id

    def test_one_raises_when_no_result(self, engine: Engine):
        """Test one raises NoResultFound when no match."""
        with Session(engine) as session:
            with pytest.raises(NoResultFound):
                session.one(Author, name="Nonexistent")

    def test_one_raises_when_multiple_results(self, engine: Engine):
        """Test one raises when multiple results found."""
        from sqlalchemy.exc import MultipleResultsFound

        with Session(engine) as session:
            author1 = Author(name="Multiple Test", field="Physics")
            author2 = Author(name="Multiple Test", field="Biology")
            session.add_all([author1, author2])
            session.flush()

            with pytest.raises(MultipleResultsFound):
                session.one(Author, name="Multiple Test")

    def test_one_or_none_success(self, engine: Engine):
        """Test one_or_none returns entity when found."""
        with Session(engine) as session:
            author = Author(name="One Or None Test", field="Chemistry")
            session.add(author)
            session.flush()
            author.revalidate()

            result = session.one_or_none(Author, name="One Or None Test")
            assert result is not None
            assert result.id == author.id

    def test_one_or_none_returns_none(self, engine: Engine):
        """Test one_or_none returns None when not found."""
        with Session(engine) as session:
            result = session.one_or_none(Author, name="Nonexistent")
            assert result is None

    def test_one_or_none_with_filters_and_expressions(self, engine: Engine):
        """Test one_or_none with both filters and expressions."""
        with Session(engine) as session:
            author = Author(name="Combined Test", field="Physics")
            session.add(author)
            session.flush()
            author.revalidate()

            result = session.one_or_none(
                Author, expressions=[Author["field"] == "Physics"], name="Combined Test"
            )
            assert result is not None
            assert result.id == author.id

    def test_first_returns_first_result(self, engine: Engine):
        """Test first returns first entity."""
        with Session(engine) as session:
            author1 = Author(name="First Test A", field="Physics")
            author2 = Author(name="First Test B", field="Physics")
            session.add_all([author1, author2])
            session.flush()
            author1.revalidate()
            author2.revalidate()

            result = session.first(Author, name="First Test A")
            assert result is not None
            assert result.id == author1.id

    def test_first_with_ordering(self, engine: Engine):
        """Test first respects order_by."""
        with Session(engine) as session:
            author1 = Author(name="Z Author", field="Physics")
            author2 = Author(name="A Author", field="Physics")
            session.add_all([author1, author2])
            session.flush()
            author1.revalidate()
            author2.revalidate()

            # Order by name ascending
            result = session.first(Author, order_bys=[Author["name"]], field="Physics")
            assert result is not None
            assert result.name == "A Author"

    def test_first_returns_none_when_no_match(self, engine: Engine):
        """Test first returns None when no results."""
        with Session(engine) as session:
            result = session.first(Author, name="Nonexistent")
            assert result is None

    def test_bulk_single_pk(self, engine: Engine):
        """Test bulk retrieval with single primary key."""
        with Session(engine) as session:
            authors = [Author(name=f"Bulk Test {i}", field="Physics") for i in range(5)]
            session.add_all(authors)
            session.flush()
            for a in authors:
                a.revalidate()

            ids = [a.id for a in authors]
            results = session.bulk(Author, ids)

            assert len(results) == 5
            assert all(r is not None for r in results)
            assert set(r.id for r in results if r) == set(ids)

    def test_bulk_preserves_order(self, engine: Engine):
        """Test bulk returns results in same order as idents."""
        with Session(engine) as session:
            authors = [
                Author(name=f"Order Test {i}", field="Physics") for i in range(3)
            ]
            session.add_all(authors)
            session.flush()
            for a in authors:
                a.revalidate()

            ids = [a.id for a in authors]
            # Request in reverse order
            reversed_ids = list(reversed(ids))
            results = session.bulk(Author, reversed_ids)

            assert [r.id for r in results if r] == reversed_ids

    def test_bulk_with_missing_ids(self, engine: Engine):
        """Test bulk handles missing IDs with None."""
        with Session(engine) as session:
            author = Author(name="Bulk Missing Test", field="Biology")
            session.add(author)
            session.flush()
            author.revalidate()

            # Request existing and non-existing IDs
            ids = [author.id, 999999, 888888]
            results = session.bulk(Author, ids)

            assert len(results) == 3
            assert results[0] is not None
            assert results[0].id == author.id
            assert results[1] is None
            assert results[2] is None

    def test_bulk_empty_list(self, engine: Engine):
        """Test bulk with empty list returns empty list."""
        with Session(engine) as session:
            results = session.bulk(Author, [])
            assert results == []

    def test_bulk_composite_pk(self, engine: Engine):
        """Test bulk retrieval with composite primary key."""
        with Session(engine) as session:
            # Create authors and publishers first
            author = Author(name="Bulk Comp Author", field="Physics")
            publisher = Publisher(name="Bulk Comp Pub", country="USA")
            session.add_all([author, publisher])
            session.flush()
            author.revalidate()
            publisher.revalidate()

            # Create books and categories
            book1 = Book(title="Book 1", year=2020)
            book1.author.value = author
            book1.publisher.value = publisher
            book2 = Book(title="Book 2", year=2021)
            book2.author.value = author
            book2.publisher.value = publisher
            book3 = Book(title="Book 3", year=2022)
            book3.author.value = author
            book3.publisher.value = publisher
            cat1 = Category(name="Category 1", description="Desc 1")
            cat2 = Category(name="Category 2", description="Desc 2")
            session.add_all([book1, book2, book3, cat1, cat2])
            session.flush()
            book1.revalidate()
            book2.revalidate()
            book3.revalidate()
            cat1.revalidate()
            cat2.revalidate()

            # Create associations via relationships
            book1.categories.append(cat1)
            book2.categories.append(cat1)
            book2.categories.append(cat2)
            session.flush()

            # Bulk retrieve with composite PK tuples
            idents = [
                (book1.id, cat1.id),
                (book2.id, cat1.id),
                (book2.id, cat2.id),
            ]
            results = session.bulk(BookCategory, idents)

            assert len(results) == 3
            assert all(r is not None for r in results)
            # Verify each result matches the expected ident
            assert results[0]
            assert results[0].book_id == book1.id
            assert results[0].category_id == cat1.id
            assert results[1]
            assert results[1].book_id == book2.id
            assert results[1].category_id == cat1.id
            assert results[2]
            assert results[2].book_id == book2.id
            assert results[2].category_id == cat2.id

    def test_bulk_composite_pk_with_missing(self, engine: Engine):
        """Test bulk with composite PK handles missing entries."""
        with Session(engine) as session:
            # Create author and publisher first
            author = Author(name="Bulk Comp Missing Author", field="Biology")
            publisher = Publisher(name="Bulk Comp Missing Pub", country="UK")
            session.add_all([author, publisher])
            session.flush()
            author.revalidate()
            publisher.revalidate()

            # Create one book and category
            book = Book(title="Bulk Comp Missing", year=2023)
            book.author.value = author
            book.publisher.value = publisher
            cat = Category(name="Bulk Comp Cat", description="Desc")
            session.add_all([book, cat])
            session.flush()
            book.revalidate()
            cat.revalidate()

            # Create association via relationship
            book.categories.append(cat)
            session.flush()

            # Request existing and non-existing composite PKs
            idents = [
                (book.id, cat.id),  # exists
                (999999, 888888),  # doesn't exist
                (777777, 666666),  # doesn't exist
            ]
            results = session.bulk(BookCategory, idents)

            assert len(results) == 3
            assert results[0] is not None
            assert results[0].book_id == book.id
            assert results[0].category_id == cat.id
            assert results[1] is None
            assert results[2] is None

    def test_count_all(self, engine: Engine):
        """Test count without filters."""
        with Session(engine) as session:
            authors = [
                Author(name=f"Count Test {i}", field="Physics") for i in range(7)
            ]
            session.add_all(authors)
            session.flush()

            count = session.count(Author, field="Physics")
            assert count >= 7  # At least the ones we added

    def test_count_with_filters(self, engine: Engine):
        """Test count with filter_by."""
        with Session(engine) as session:
            author1 = Author(name="Count Filter 1", field="Physics")
            author2 = Author(name="Count Filter 2", field="Biology")
            author3 = Author(name="Count Filter 3", field="Physics")
            session.add_all([author1, author2, author3])
            session.flush()

            count_physics = session.count(Author, field="Physics")
            count_biology = session.count(Author, field="Biology")

            assert count_physics >= 2
            assert count_biology >= 1

    def test_count_with_expressions(self, engine: Engine):
        """Test count with where expressions."""
        with Session(engine) as session:
            authors = [
                Author(name="Count Expr A", field="Physics"),
                Author(name="Count Expr B", field="Physics"),
                Author(name="Count Expr C", field="Biology"),
            ]
            session.add_all(authors)
            session.flush()

            count = session.count(
                Author,
                expressions=[
                    Author["name"].like("Count Expr%"),
                    Author["field"] == "Physics",
                ],
            )
            assert count >= 2

    def test_list_basic(self, engine: Engine):
        """Test list returns multiple entities."""
        with Session(engine) as session:
            authors = [
                Author(name=f"List Test {i}", field="Chemistry") for i in range(5)
            ]
            session.add_all(authors)
            session.flush()
            for a in authors:
                a.revalidate()

            results = session.list(Author, field="Chemistry")
            assert len(results) >= 5

    def test_list_with_limit(self, engine: Engine):
        """Test list respects limit."""
        with Session(engine) as session:
            authors = [
                Author(name=f"Limit Test {i}", field="Chemistry") for i in range(10)
            ]
            session.add_all(authors)
            session.flush()

            results = session.list(Author, limit=3, field="Chemistry")
            assert len(results) == 3

    def test_list_with_offset(self, engine: Engine):
        """Test list respects offset."""
        with Session(engine) as session:
            authors = [
                Author(name=f"Offset Test {i}", field="Literature") for i in range(5)
            ]
            session.add_all(authors)
            session.flush()
            for a in authors:
                a.revalidate()

            # Get all, then with offset
            all_results = session.list(
                Author, limit=100, order_bys=[Author["id"]], field="Literature"
            )
            offset_results = session.list(
                Author,
                limit=100,
                offset=2,
                order_bys=[Author["id"]],
                field="Literature",
            )

            assert len(all_results) >= 5
            assert len(offset_results) >= 3
            assert offset_results[0].id == all_results[2].id

    def test_list_with_ordering(self, engine: Engine):
        """Test list respects order_by."""
        with Session(engine) as session:
            author1 = Author(name="Z List", field="History")
            author2 = Author(name="A List", field="History")
            author3 = Author(name="M List", field="History")
            session.add_all([author1, author2, author3])
            session.flush()

            results = session.list(Author, order_bys=[Author["name"]], field="History")
            names = [r.name for r in results]
            # Should be in ascending order
            assert names.index("A List") < names.index("M List") < names.index("Z List")

    def test_list_with_filters_and_expressions(self, engine: Engine):
        """Test list with both filters and expressions."""
        with Session(engine) as session:
            authors = [
                Author(name="List Filter A", field="Physics"),
                Author(name="List Filter B", field="Physics"),
                Author(name="List Filter C", field="Biology"),
            ]
            session.add_all(authors)
            session.flush()

            results = session.list(
                Author,
                expressions=[Author["name"].like("List Filter%")],
                field="Physics",
            )
            assert len(results) >= 2
            assert all(r.field == "Physics" for r in results)

    def test_partitions_yields_chunks(self, engine: Engine):
        """Test partitions yields results in chunks."""
        with Session(engine) as session:
            authors = [
                Author(name=f"Partition Test {i}", field="Astronomy") for i in range(15)
            ]
            session.add_all(authors)
            session.flush()

            partitions_list = list(
                session.partitions(Author, size=5, limit=15, field="Astronomy")
            )

            # Should have 3 partitions of size 5
            assert len(partitions_list) >= 3
            for partition in partitions_list[:3]:
                assert len(partition) <= 5

    def test_partitions_respects_limit(self, engine: Engine):
        """Test partitions respects total limit."""
        with Session(engine) as session:
            authors = [
                Author(name=f"Part Limit Test {i}", field="History") for i in range(20)
            ]
            session.add_all(authors)
            session.flush()

            all_results = []
            for partition in session.partitions(
                Author, size=3, limit=7, field="History"
            ):
                all_results.extend(partition)

            assert len(all_results) == 7

    def test_partitions_with_ordering(self, engine: Engine):
        """Test partitions respects ordering."""
        with Session(engine) as session:
            authors = [
                Author(name=f"Part Order {chr(65 + i)}", field="Literature")
                for i in range(10)
            ]
            session.add_all(authors)
            session.flush()

            all_results = []
            for partition in session.partitions(
                Author, size=3, order_bys=[Author["name"]], field="Literature"
            ):
                all_results.extend(partition)

            names = [r.name for r in all_results]
            # Should be in ascending order
            assert names == sorted(names)

    def test_partitions_with_filters(self, engine: Engine):
        """Test partitions with filters."""
        with Session(engine) as session:
            authors = [
                Author(name="Part Filter Physics 1", field="Physics"),
                Author(name="Part Filter Physics 2", field="Physics"),
                Author(name="Part Filter Biology 1", field="Biology"),
            ]
            session.add_all(authors)
            session.flush()

            all_results = []
            for partition in session.partitions(Author, size=2, field="Physics"):
                all_results.extend(partition)

            assert len(all_results) >= 2
            assert all(r.field == "Physics" for r in all_results)


class TestObjectIdentity:
    """Test object identity and caching within a Session."""

    def test_identity_map_same_object(self, engine: Engine):
        """Test that fetching the same PK returns the same Python object."""
        with Session(engine) as session:
            author = Author(name="Identity Test", field="Biology")
            session.add(author)
            session.flush()
            author.revalidate()

            # Fetch twice
            fetch1 = session.get_one(Author, author.id)
            fetch2 = session.get_one(Author, author.id)

            # Should be the same object
            assert fetch1 is fetch2

    def test_identity_across_queries(self, engine: Engine):
        """Test identity is maintained across different query methods."""
        with Session(engine) as session:
            author = Author(name="Query Identity Test", field="Chemistry")
            session.add(author)
            session.flush()
            author.revalidate()

            # Fetch via get_one
            via_get = session.get_one(Author, author.id)

            # Fetch via query
            stmt = select(Author).where(Author["id"] == author.id)
            via_query = session.execute(stmt).scalars().one()

            # Should be the same object
            assert via_get is via_query

    def test_identity_not_shared_across_sessions(self, engine: Engine):
        """Test that different sessions have different object instances."""
        with Session(engine) as session1:
            author = Author(name="Cross Session Test", field="Physics")
            session1.add(author)
            session1.flush()
            author.revalidate()
            session1.commit()
            obj1 = session1.get_one(Author, author.id)

            with Session(engine) as session2:
                obj2 = session2.get_one(Author, author.id)

                # Different objects
                assert obj1 is not obj2
                # But same data
                assert obj1.name == obj2.name

    def test_modified_object_reflects_in_identity_map(self, engine: Engine):
        """Test that modifying an object is visible through identity map."""
        with Session(engine) as session:
            author = Author(name="Original", field="Biology")
            session.add(author)
            session.flush()
            author.revalidate()

            # Get reference
            ref1 = session.get_one(Author, author.id)

            # Modify through first reference
            ref1.name = "Modified"

            # Get another reference
            ref2 = session.get_one(Author, author.id)

            # Should see modification
            assert ref2.name == "Modified"
            assert ref1 is ref2


class TestAddOperations:
    """Test Session.add and add_all operations."""

    def test_add_single_object(self, engine: Engine):
        """Test adding a single object."""
        with Session(engine) as session:
            author = Author(name="Single Add", field="Literature")
            session.add(author)
            session.flush()
            author.revalidate()

            assert author.id is not None

    def test_add_all_multiple_objects(self, engine: Engine):
        """Test adding multiple objects at once."""
        with Session(engine) as session:
            authors = [Author(name=f"Batch {i}", field="Physics") for i in range(5)]
            session.add_all(authors)
            session.flush()
            for a in authors:
                a.revalidate()

            assert all(a.id is not None for a in authors)

    def test_add_with_relationships(self, engine: Engine):
        """Test that adding parent adds related children."""
        with Session(engine) as session:
            author = Author(name="Parent Author", field="Biology")
            publisher = Publisher(name="Parent Publisher", country="USA")

            book = Book(title="Child Book", year=2024)
            book.author.value = author
            book.publisher.value = publisher

            # Adding book should cascade
            session.add(book)
            session.flush()
            book.revalidate()
            author.revalidate()
            publisher.revalidate()

            assert book.id is not None
            assert author.id is not None
            assert publisher.id is not None


class TestDeleteOperations:
    """Test Session.delete operations."""

    def test_delete_single_object(self, engine: Engine):
        """Test deleting a single object."""
        with Session(engine) as session:
            author = Author(name="To Delete", field="Chemistry")
            session.add(author)
            session.flush()
            author.revalidate()

            # Delete
            session.delete(author)
            session.flush()

            # Verify
            assert session.get(Author, author.id) is None

    def test_delete_cascades_to_children(self, engine: Engine):
        """Test that deleting parent cascades to children."""
        with Session(engine) as session:
            author = Author(name="Cascade Author", field="Physics")
            publisher = Publisher(name="Cascade Publisher", country="USA")

            book = Book(title="Cascade Book", year=2024)
            book.author.value = author
            book.publisher.value = publisher

            session.add(book)
            session.flush()
            book.revalidate()
            author.revalidate()

            # Delete author (should cascade to book due to ON DELETE CASCADE)
            session.delete(author)
            session.flush()

            # Book should be deleted
            assert session.get(Book, book.id) is None
            assert session.get(Author, author.id) is None


class TestFlushCommitRollback:
    """Test flush, commit, and rollback operations."""

    def test_flush_makes_changes_visible_in_session(self, engine: Engine):
        """Test that flush makes changes visible within the session."""
        with Session(engine) as session:
            author = Author(name="Flush Test", field="Biology")
            session.add(author)

            # Before flush, no ID
            assert author.id is None

            # After flush
            session.flush()
            author.revalidate()
            assert author.id is not None

            # But not yet committed
            # (We can't easily test this in the same session)

    def test_commit_persists_changes(self, engine: Engine):
        """Test that commit persists changes to database."""
        with Session(engine) as session:
            author = Author(name="Commit Test", field="Chemistry")
            session.add(author)
            session.flush()
            author.revalidate()
            author_id = author.id
            session.commit()

        # Verify in new session
        with Session(engine) as session:
            retrieved = session.get_one(Author, author_id)
            assert retrieved.name == "Commit Test"

    def test_rollback_reverts_changes(self, engine: Engine):
        """Test that rollback reverts unflushed/uncommitted changes."""
        with Session(engine) as session:
            author = Author(name="Rollback Test", field="Physics")
            session.add(author)
            session.flush()
            author.revalidate()
            session.commit()  # Commit the original insert

            # Modify
            author.name = "Modified"
            session.flush()

            # Rollback only reverts the modification
            session.rollback()

            # Refresh the author object to reload from database
            session.refresh(author)
            author.revalidate()
            assert author.name == "Rollback Test"

    def test_multiple_flush_before_commit(self, engine: Engine):
        """Test multiple flushes before final commit."""
        with Session(engine) as session:
            author = Author(name="Multi Flush", field="Literature")
            session.add(author)
            session.flush()

            author.name = "Multi Flush Updated 1"
            session.flush()

            author.name = "Multi Flush Updated 2"
            session.flush()

            assert author.name == "Multi Flush Updated 2"


class TestExpireRefresh:
    """Test expire and refresh operations."""

    def test_expire_reloads_on_access(self, engine: Engine):
        """Test that expiring an object reloads it on next access."""
        with Session(engine) as session:
            author = Author(name="Expire Test", field="Biology")
            session.add(author)
            session.flush()
            author.revalidate()

            # Modify in database directly (simulate external change)
            stmt = (
                update(Author)
                .where(Author["id"] == author.id)
                .values(name="Externally Modified")
            )
            session.execute(stmt)
            session.commit()

            # Object still has old value in memory
            # (Unless we expire and refresh it)

            # Expire and refresh to reload from database
            session.expire(author)
            session.refresh(author)
            author.revalidate()
            # Next access should show new value
            assert author.name == "Externally Modified"

    def test_refresh_reloads_immediately(self, engine: Engine):
        """Test that refresh reloads object immediately."""
        with Session(engine) as session:
            author = Author(name="Refresh Test", field="Chemistry")
            session.add(author)
            session.flush()
            author.revalidate()

            # Modify in database
            stmt = (
                update(Author)
                .where(Author["id"] == author.id)
                .values(name="DB Modified")
            )
            session.execute(stmt)
            session.commit()

            # Refresh to load new value
            session.refresh(author)
            assert author.name == "DB Modified"

    def test_expire_all(self, engine: Engine):
        """Test expiring all objects in session."""
        with Session(engine) as session:
            author1 = Author(name="Expire All 1", field="Physics")
            author2 = Author(name="Expire All 2", field="Biology")
            session.add_all([author1, author2])
            session.flush()

            # Expire all
            session.expire_all()

            # Next access should reload
            assert author1.name == "Expire All 1"
            assert author2.name == "Expire All 2"


class TestExpunge:
    """Test expunging objects from session."""

    def test_expunge_removes_from_session(self, engine: Engine):
        """Test that expunge removes object from session."""
        with Session(engine) as session:
            author = Author(name="Expunge Test", field="Literature")
            session.add(author)
            session.flush()
            author.revalidate()

            # Expunge
            session.expunge(author)

            # Object no longer in session
            # Getting it again returns a different instance
            author2 = session.get_one(Author, author.id)
            assert author is not author2

    def test_expunge_all(self, engine: Engine):
        """Test expunging all objects from session."""
        with Session(engine) as session:
            author1 = Author(name="Expunge All 1", field="Physics")
            author2 = Author(name="Expunge All 2", field="Biology")
            session.add_all([author1, author2])
            session.flush()
            author1.revalidate()
            author2.revalidate()

            # Expunge all
            session.expunge_all()

            # Fetching again returns new instances
            new1 = session.get_one(Author, author1.id)
            new2 = session.get_one(Author, author2.id)

            assert new1 is not author1
            assert new2 is not author2


class TestMerge:
    """Test Session.merge functionality."""

    def test_merge_detached_object(self, engine: Engine, test_id: uuid.UUID):
        """Test merging a detached object back into a session."""
        author_id = None
        author_detached = None

        # Create and detach
        with Session(engine) as session1:
            author = Author(name="Merge Test", field="Physics", test_id=test_id)
            session1.add(author)
            session1.flush()
            author.revalidate()
            session1.commit()
            author_id = author.id
            author_detached = author

        # Author is now detached

        # Merge into new session
        with Session(engine) as session2:
            author_merged = session2.merge(author_detached)

            # Verify the merged object is revalidated and works correctly
            assert author_merged.id == author_id
            assert author_merged.name == "Merge Test"
            assert author_merged.field == "Physics"

            # Modify and flush to verify it's properly tracked
            author_merged.name = "Merged and Modified"
            session2.flush()
            session2.commit()

        # Verify changes persisted
        with Session(engine) as session3:
            author_final = session3.get_one(Author, author_id)
            assert author_final.name == "Merged and Modified"

    def test_merge_with_modifications(self, engine: Engine):
        """Test merging an object with pending modifications."""
        # Create original
        with Session(engine) as session1:
            author = Author(name="Original Name", field="Biology")
            session1.add(author)
            session1.flush()
            author.revalidate()
            session1.commit()
            author_id = author.id

        # Get in new session and modify
        with Session(engine) as session2:
            author = session2.get_one(Author, author_id)

        # Now author is detached, modify it
        author.name = "Modified While Detached"

        # Merge back into another session
        with Session(engine) as session3:
            merged = session3.merge(author)
            assert merged.name == "Modified While Detached"
            session3.commit()

        # Verify persistence
        with Session(engine) as session4:
            final = session4.get_one(Author, author_id)
            assert final.name == "Modified While Detached"

    def test_merge_with_load_false(self, engine: Engine):
        """Test merge with load=False parameter."""
        # Create object
        with Session(engine) as session1:
            author = Author(name="Load Test", field="Chemistry")
            session1.add(author)
            session1.flush()
            author.revalidate()
            session1.commit()
            author_id = author.id

        # Get and detach
        with Session(engine) as session2:
            author = session2.get_one(Author, author_id)

        # Merge with load=False
        with Session(engine) as session3:
            merged = session3.merge(author, load=False)
            assert merged.id == author_id
            assert merged.name == "Load Test"

    def test_merge_updates_existing_in_session(self, engine: Engine):
        """Test that merge updates an already-loaded instance."""
        author_id = None

        # Create
        with Session(engine) as session1:
            author = Author(name="Initial Name", field="History")
            session1.add(author)
            session1.flush()
            author.revalidate()
            session1.commit()
            author_id = author.id

        with Session(engine) as session2:
            # Load into session
            author_in_session = session2.get_one(Author, author_id)
            assert author_in_session.name == "Initial Name"

            # Create a detached copy with different data
            author_detached = Author(id=author_id, name="Updated Name", field="History")

            # Merge should update the existing instance
            merged = session2.merge(author_detached)

            # The merged instance should be the same object that was already in session
            assert merged is author_in_session
            assert merged.name == "Updated Name"

    def test_merge_returns_revalidated_transmuter(self, engine: Engine):
        """Test that merge returns a properly revalidated transmuter instance."""
        # Create
        with Session(engine) as session1:
            publisher = Publisher(name="Test Publisher", country="USA")
            session1.add(publisher)
            session1.flush()
            publisher.revalidate()
            session1.commit()
            publisher_id = publisher.id

        # Detach
        with Session(engine) as session2:
            publisher = session2.get_one(Publisher, publisher_id)

        # Merge and verify it's a transmuter with proper proxy
        with Session(engine) as session3:
            merged = session3.merge(publisher)

            # Should be a transmuter instance
            assert isinstance(merged, Publisher)
            assert isinstance(merged, BaseTransmuter)

            # Should have proper ORM backing
            assert merged.__transmuter_provided__ is not None

            # Modify to ensure it's properly tracked
            merged.name = "Updated Publisher"
            session3.flush()
            session3.commit()

        # Verify update persisted
        with Session(engine) as session4:
            final = session4.get_one(Publisher, publisher_id)
            assert final.name == "Updated Publisher"

    def test_merge_maintains_validation_context(self, engine: Engine):
        """Test that merge properly maintains validation context."""
        # Create a simple object
        with Session(engine) as session1:
            author = Author(name="Context Author", field="Physics")
            session1.add(author)
            session1.flush()
            author.revalidate()
            session1.commit()
            author_id = author.id

        # Load and detach
        with Session(engine) as session2:
            author = session2.get_one(Author, author_id)

        # Merge into new session
        with Session(engine) as session3:
            merged_author = session3.merge(author)

            # The merged object should be in the validation context
            assert merged_author.__transmuter_provided__ in session3._validation_context
            assert (
                session3._validation_context[merged_author.__transmuter_provided__]
                is merged_author
            )

    def test_merge_with_multiple_objects(self, engine: Engine):
        """Test merging multiple objects."""
        # Create multiple objects
        with Session(engine) as session1:
            author1 = Author(name="Author 1", field="Biology")
            author2 = Author(name="Author 2", field="Chemistry")

            session1.add_all([author1, author2])
            session1.flush()
            author1.revalidate()
            author2.revalidate()
            session1.commit()

            author1_id = author1.id
            author2_id = author2.id

        # Load and detach
        with Session(engine) as session2:
            author1 = session2.get_one(Author, author1_id)
            author2 = session2.get_one(Author, author2_id)

        # Modify while detached
        author1.name = "Modified Author 1"
        author2.field = "Astronomy"

        # Merge both into new session
        with Session(engine) as session3:
            merged1 = session3.merge(author1)
            merged2 = session3.merge(author2)

            # Verify modifications are present
            assert merged1.name == "Modified Author 1"
            assert merged2.field == "Astronomy"

            session3.flush()
            session3.commit()

        # Verify all changes persisted
        with Session(engine) as session4:
            final1 = session4.get_one(Author, author1_id)
            final2 = session4.get_one(Author, author2_id)

            assert final1.name == "Modified Author 1"
            assert final2.field == "Astronomy"

    def test_merge_after_expunge(self, engine: Engine):
        """Test merging an object that was explicitly expunged."""
        with Session(engine) as session1:
            author = Author(name="Expunge Test", field="History")
            session1.add(author)
            session1.flush()
            author.revalidate()
            author_id = author.id

            # Expunge the object
            session1.expunge(author)

            # Merge it back
            merged = session1.merge(author)

            assert merged.id == author_id
            assert merged.name == "Expunge Test"

            # Modify and commit
            merged.field = "Literature"
            session1.commit()

        # Verify changes persisted
        with Session(engine) as session2:
            final = session2.get_one(Author, author_id)
            assert final.field == "Literature"

    def test_merge_revalidates_transmuter(self, engine: Engine):
        """Test that merge properly revalidates the transmuter instance."""
        # Create
        with Session(engine) as session1:
            category = Category(name="Test Category", description="Test Description")
            session1.add(category)
            session1.flush()
            category.revalidate()
            session1.commit()
            category_id = category.id

        # Load and detach
        with Session(engine) as session2:
            category = session2.get_one(Category, category_id)

        # Verify it's detached (no longer in session)
        # Merge into new session
        with Session(engine) as session3:
            merged = session3.merge(category)

            # Verify returned instance is properly setup
            assert isinstance(merged, Category)
            assert isinstance(merged, Transmuter)
            assert merged.id == category_id
            assert merged.name == "Test Category"

            # Modify to confirm tracking
            merged.description = "Updated Description"
            session3.flush()
            session3.commit()

        # Verify update
        with Session(engine) as session4:
            final = session4.get_one(Category, category_id)
            assert final.description == "Updated Description"



"""Test Transmuter-ORM integration.

Tests:
- ORM object creation during transmuter initialization
- ORM object creation during validation
- Blessing rules and constraints
- from_attributes and model_validate
"""

from __future__ import annotations

import pytest
from pydantic import ValidationError
from sqlalchemy import Engine

from arcanus.materia.sqlalchemy import Session
from arcanus.materia.sqlalchemy.base import SqlalchemyMateria
from tests import models
from tests.transmuters import Author, Book, Publisher


class TestTransmuterORMCreation:
    """Test that transmuters create and manage ORM objects correctly."""

    def test_transmuter_init_creates_orm(self):
        """Test that initializing a transmuter creates an underlying ORM object."""
        author = Author(name="Isaac Asimov", field="Literature")

        assert author.__transmuter_provided__ is not None
        assert isinstance(author.__transmuter_provided__, models.Author)
        assert author.__transmuter_provided__.name == "Isaac Asimov"
        assert author.__transmuter_provided__.field == "Literature"

    def test_transmuter_validation_creates_orm(self):
        """Test that model_validate creates ORM objects."""
        data = {"name": "Arthur C. Clarke", "field": "Literature"}
        author = Author.model_validate(data)

        assert author.__transmuter_provided__ is not None
        assert isinstance(author.__transmuter_provided__, models.Author)
        assert author.name == "Arthur C. Clarke"

    def test_from_attributes_converts_orm_to_transmuter(self, engine: Engine):
        """Test that from_attributes=True allows ORM -> Transmuter conversion."""
        # Create ORM object directly using native SQLAlchemy session
        from sqlalchemy.orm import Session as SASession

        with SASession(engine) as sa_session:
            orm_author = models.Author(name="Philip K. Dick", field="Literature")
            sa_session.add(orm_author)
            sa_session.flush()
            sa_session.commit()
            sa_session.refresh(orm_author)

            # Convert ORM to Transmuter
            author = Author.model_validate(orm_author)

            assert isinstance(author, Author)
            assert author.name == "Philip K. Dick"
            assert author.__transmuter_provided__ is orm_author

    def test_identity_field_readonly_after_set(self, engine: Engine):
        """Test that identity fields are frozen after being set."""
        with Session(engine) as session:
            author = Author(name="Test Author", field="Physics")
            session.add(author)
            session.flush()
            author.revalidate()

            # ID is now set
            assert author.id is not None

            # Try to modify ID - should raise ValidationError
            with pytest.raises(ValidationError, match="frozen"):
                author.id = 999

    def test_transmuter_without_id_can_be_persisted(self, engine: Engine):
        """Test creating a transmuter without ID and persisting it."""
        author = Author(name="New Author", field="Biology")

        assert author.id is None
        assert author.__transmuter_provided__ is not None

        with Session(engine) as session:
            session.add(author)
            session.flush()
            author.revalidate()

            # ID should now be set
            assert author.id is not None
            assert author.__transmuter_provided__.id is not None  # type: ignore

    def test_nested_transmuter_creation(self, engine: Engine):
        """Test creating nested transmuters with relationships."""
        publisher = Publisher(name="Test Publisher", country="USA")
        author = Author(name="Test Author", field="Chemistry")

        # Create book with relationships
        book = Book(
            title="Test Book",
            year=2024,
        )
        book.author.value = author
        book.publisher.value = publisher

        # All should have ORM objects
        assert book.__transmuter_provided__ is not None
        assert author.__transmuter_provided__ is not None
        assert publisher.__transmuter_provided__ is not None

        # Persist
        with Session(engine) as session:
            session.add(book)
            session.flush()
            book.revalidate()
            author.revalidate()
            publisher.revalidate()

            assert book.id is not None
            assert author.id is not None
            assert publisher.id is not None


class TestBlessingRules:
    """Test transmuter blessing constraints and rules."""

    def test_transmuter_blessed_once(self):
        """Test that a transmuter class can only be blessed with one ORM type."""
        from arcanus.base import BaseTransmuter

        # This should work - fresh materia
        test_materia = SqlalchemyMateria()

        @test_materia.bless(models.Author)
        class TestAuthor(BaseTransmuter):
            id: int | None = None
            name: str
            field: str

        # Attempting to bless same class again should fail
        with pytest.raises(RuntimeError, match="already blessed"):
            test_materia.bless(models.Publisher)(TestAuthor)  # Different ORM class

    def test_different_transmuters_same_orm(self):
        """Test that different transmuter classes can be blessed with the same ORM."""
        from arcanus.base import BaseTransmuter

        test_materia = SqlalchemyMateria()

        @test_materia.bless(models.Author)
        class AuthorV1(BaseTransmuter):
            id: int | None = None
            name: str
            field: str

        # This should work - different transmuter class, same ORM
        @test_materia.bless(models.Author)
        class AuthorV2(BaseTransmuter):
            id: int | None = None
            name: str
            field: str

        with test_materia:
            # Both should work independently
            author1 = AuthorV1(name="Author 1", field="Physics")
            author2 = AuthorV2(name="Author 2", field="Biology")

            assert isinstance(author1.__transmuter_provided__, models.Author)
            assert isinstance(author2.__transmuter_provided__, models.Author)

    def test_unblessed_transmuter_no_orm(self):
        """Test that unblessed transmuters don't have ORM objects."""
        from arcanus.base import BaseTransmuter

        class UnblessedTransmuter(BaseTransmuter):
            name: str

        obj = UnblessedTransmuter(name="test")
        assert obj.__transmuter_provided__ is None


class TestORMAttributeAccess:
    """Test accessing ORM attributes through transmuters."""

    def test_transmuter_provides_orm_attributes(self):
        """Test that transmuter fields map to ORM attributes."""
        author = Author(name="Test", field="Physics")

        assert author.name == author.__transmuter_provided__.name  # type: ignore
        assert author.field == author.__transmuter_provided__.field  # type: ignore

    def test_modify_transmuter_updates_orm(self):
        """Test that modifying transmuter fields updates ORM object."""
        author = Author(name="Original", field="Physics")

        # Modify through transmuter
        author.name = "Updated"

        # ORM should reflect change
        assert author.__transmuter_provided__.name == "Updated"  # type: ignore

    def test_modify_orm_visible_in_transmuter(self):
        """Test that modifying ORM attributes is visible through transmuter after revalidate."""
        author = Author(name="Original", field="Physics")

        # Modify ORM directly
        author.__transmuter_provided__.name = "Changed"  # type: ignore

        # Need to revalidate to sync from ORM
        author.revalidate()

        # Should be visible through transmuter
        assert author.name == "Changed"

    def test_orm_and_transmuter_stay_in_sync(self, engine: Engine):
        """Test that ORM and transmuter stay synchronized during persistence."""
        with Session(engine) as session:
            author = Author(name="Sync Test", field="Biology")
            session.add(author)

            # Modify before flush
            author.name = "Updated Before Flush"
            session.flush()

            assert author.__transmuter_provided__.name == "Updated Before Flush"  # type: ignore

            # Modify after flush
            author.name = "Updated After Flush"
            session.flush()

            assert author.__transmuter_provided__.name == "Updated After Flush"  # type: ignore



from arcanus.association import (
    Relation,
    RelationCollection,
    RelationMap,
    RelationMaps,
    RelationSet,
    Relationship,
    Relationships,
    TypedRelationMap,
    TypedRelationMaps,
)
from arcanus.base import (
    BaseTransmuter,
    Transmuter,
)

__all__ = [
    "BaseTransmuter",
    "Transmuter",
    "Relation",
    "RelationCollection",
    "RelationMap",
    "RelationMaps",
    "RelationSet",
    "Relationship",
    "Relationships",
    "TypedRelationMap",
    "TypedRelationMaps",
]



from __future__ import annotations

import pytest
from pydantic import ValidationError

from tests.transmuters import Author, Book

"""Test transmuter-specific features with NoOpMateria.

This module tests:
- Create/Update model generation and usage
- shell() and absorb() methods
- Validation context
- Additional properties and custom behavior
- Model inheritance and metaclass features
"""


class TestCreateModel:
    """Test the dynamically generated Create model."""

    def test_create_model_excludes_identity_fields(self):
        """Test that Create model excludes Identity fields."""
        # Create model should not have 'id' field
        partial = Book.Create(title="New Book", year=2024)

        assert hasattr(partial, "title")
        assert hasattr(partial, "year")
        assert not hasattr(partial, "id")  # Identity field excluded

    def test_create_model_excludes_associations(self):
        """Test that Create model excludes association fields."""
        partial = Book.Create(title="New Book", year=2024)

        # Association fields should not be in Create model
        assert not hasattr(partial, "author")
        assert not hasattr(partial, "publisher")
        assert not hasattr(partial, "categories")

    def test_create_model_with_valid_data(self):
        """Test creating valid Create model instances."""
        partial = Book.Create(
            title="Test Book",
            year=2024,
            author_id=1,
            publisher_id=2,
        )

        data = partial.model_dump(exclude_none=True)
        assert data["title"] == "Test Book"
        assert data["year"] == 2024
        assert data["author_id"] == 1
        assert data["publisher_id"] == 2

    def test_create_model_requires_non_optional_fields(self):
        """Test that Create model still validates required fields."""
        with pytest.raises(ValidationError) as exc_info:
            Book.Create(year=2024)  # Missing required 'title'

        errors = exc_info.value.errors()
        assert any("title" in str(error["loc"]) for error in errors)

    def test_create_model_name_and_module(self):
        """Test that Create model has proper naming."""
        assert Book.Create.__name__ == "BookCreate"
        assert Author.Create.__name__ == "AuthorCreate"


class TestUpdateModel:
    """Test the dynamically generated Update model."""

    def test_update_model_makes_all_fields_optional(self):
        """Test that Update model makes all non-frozen fields optional."""
        partial = Book.Update(title="Updated Title")

        data = partial.model_dump(exclude_none=True)
        assert "title" in data
        assert data["title"] == "Updated Title"

        # Other fields should have None as default
        assert "year" not in data
        assert "author_id" not in data

    def test_update_model_excludes_associations(self):
        """Test that Update model excludes association fields."""
        partial = Book.Update(title="Updated")

        assert not hasattr(partial, "author")
        assert not hasattr(partial, "publisher")
        assert not hasattr(partial, "categories")

    def test_update_model_with_multiple_fields(self):
        """Test Update model with multiple fields."""
        partial = Book.Update(
            title="New Title",
            year=2025,
        )

        data = partial.model_dump(exclude_none=True)
        assert data["title"] == "New Title"
        assert data["year"] == 2025
        assert "author_id" not in data

    def test_update_model_with_empty_update(self):
        """Test creating Update model with no fields."""
        partial = Book.Update()

        # All fields should be None when not provided
        data = partial.model_dump(exclude_none=True)
        assert "title" not in data
        assert "year" not in data

    def test_update_model_name_and_module(self):
        """Test that Update model has proper naming."""
        assert Book.Update.__name__ == "BookUpdate"
        assert Author.Update.__name__ == "AuthorUpdate"


class TestShellMethod:
    """Test the shell() class method for creating instances from partial models."""

    def test_shell_from_create_model(self):
        """Test creating full model from Create model."""
        partial = Book.Create(title="Partial Book", year=2024)
        book = Book.shell(partial)

        assert isinstance(book, Book)
        assert book.title == "Partial Book"
        assert book.year == 2024
        assert book.id is None  # Identity field gets default value

    def test_shell_ignores_extra_fields_in_partial(self):
        """Test that shell() ignores fields that shouldn't exist in target."""
        # Create model with extra 'id' field (should be ignored)
        partial = Book.Create(
            id=999,  # This should be ignored
            title="Test Book",
            year=2024,
        )
        book = Book.shell(partial)

        # ID should be None (default), not 999
        assert book.id is None
        assert book.title == "Test Book"
        assert book.year == 2024

    def test_shell_from_dict(self):
        """Test creating model from dictionary using shell()."""
        data = {"title": "Dict Book", "year": 2023}
        book = Book.shell(Book.Create(**data))

        assert book.title == "Dict Book"
        assert book.year == 2023

    def test_shell_preserves_valid_fields(self):
        """Test that shell() preserves all valid fields from partial."""
        partial = Author.Create(name="Test Author", field="Physics")
        author = Author.shell(partial)

        assert author.name == "Test Author"
        assert author.field == "Physics"
        assert author.id is None


class TestAbsorbMethod:
    """Test the absorb() method for updating instances."""

    def test_absorb_updates_provided_fields(self):
        """Test that absorb() updates only the provided fields."""
        book = Book(id=1, title="Original Title", year=2020)
        partial = Book.Update(title="Updated Title")

        updated_book = book.absorb(partial)

        assert updated_book is book  # Should be same instance
        assert book.title == "Updated Title"
        assert book.year == 2020  # Unchanged
        assert book.id == 1  # Unchanged

    def test_absorb_ignores_none_values(self):
        """Test that absorb() behavior with None values in Update model."""
        book = Book(id=2, title="Original", year=2020, author_id=5)
        partial = Book.Update(title="Updated", author_id=None)

        book.absorb(partial)

        # Title should be updated
        assert book.title == "Updated"
        # absorb() updates with all provided values, including explicit None
        # If you want to skip None values, filter them before calling absorb
        assert book.author_id is None  # Updated to None as explicitly set

        # Test filtering None values before absorb
        book2 = Book(id=3, title="Original2", year=2021, author_id=10)
        partial2 = Book.Update(title="Updated2", author_id=None)
        # Filter out None values
        update_data = {k: v for k, v in partial2.model_dump().items() if v is not None}
        book2.absorb(Book.Update(**update_data))

        assert book2.title == "Updated2"
        assert book2.author_id == 10  # Not updated since None was filtered

    def test_absorb_with_multiple_updates(self):
        """Test absorbing multiple fields at once."""
        author = Author(id=1, name="Original Name", field="Physics")
        partial = Author.Update(name="New Name", field="Biology")

        author.absorb(partial)

        assert author.name == "New Name"
        assert author.field == "Biology"
        assert author.id == 1  # Unchanged

    def test_absorb_ignores_identity_fields(self):
        """Test that absorb() ignores identity fields even if provided."""
        book = Book(id=1, title="Test", year=2020)

        # Update model shouldn't have 'id', but if someone passes it in dict
        partial = Book.Update(id=999, title="Updated")

        book.absorb(partial)

        assert book.id == 1  # ID unchanged
        assert book.title == "Updated"

    def test_absorb_returns_self(self):
        """Test that absorb() returns the same instance."""
        author = Author(id=1, name="Test", field="Physics")
        partial = Author.Update(name="Updated")

        result = author.absorb(partial)

        assert result is author


class TestModelMetadata:
    """Test metadata and introspection of transmuter models."""

    def test_model_associations_property(self):
        """Test that model_associations returns association fields."""
        associations = Book.model_associations

        # Association fields
        assert "author" in associations
        assert "publisher" in associations
        assert "categories" in associations

        # Non-association fields should not be present
        assert "title" not in associations
        assert "year" not in associations

    def test_model_identities_property(self):
        """Test that model_identities returns identity fields."""
        identities = Book.model_identities

        assert "id" in identities

        # Non-identity fields should not be present
        assert "title" not in identities
        assert "year" not in identities

    def test_model_fields_property(self):
        """Test that model_fields contains all fields."""
        fields = Book.model_fields

        # Should contain all fields
        assert "id" in fields
        assert "title" in fields
        assert "year" in fields
        assert "author" in fields
        assert "publisher" in fields

    def test_model_config(self):
        """Test model configuration."""
        assert Book.model_config.get("from_attributes") is True
        assert Author.model_config.get("from_attributes") is True



"""
SQLAlchemy Materia Benchmark Tests - ORM CRUD Operations

This module benchmarks arcanus's SQLAlchemy materia against common patterns
for CRUD API endpoints. Comparing three approaches:

1. PURE SQLALCHEMY (Baseline)
   - Direct ORM operations without Pydantic validation

2. PYDANTIC + SQLALCHEMY (Common Pattern)
   - Validate with Pydantic first, then model_dump to create ORM objects
   - This is what most developers do today

3. arcanus (Combined)
   - Use transmuter with full Pydantic validation and SQLAlchemy integration
   - Validates and produces ORM-integrated objects in one step

Each test uses transactions that rollback to avoid polluting test data.

Run locally:  pytest benchmark/ -v --benchmark-enable
With CodSpeed: pytest benchmark/ --codspeed
"""

from __future__ import annotations

import random

import pytest
from sqlalchemy import select
from sqlalchemy.orm import joinedload, selectinload

from tests import models, schemas
from tests.transmuters import Author, Book

# Batch size for benchmark tests
BATCH_SIZE = 50


class TestCreateSingleAuthor:
    """
    Benchmark single author creation (flat object, no relationships).

    Simulates: POST /api/authors with simple payload
    Each iteration creates one author from random selection.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="create-single-author")
    def test_sqlalchemy_create(self, benchmark, session_factory, create_author_data):
        """Pure SQLAlchemy: Direct ORM creation."""
        data = random.choice(create_author_data)
        author_data = {"name": data["name"], "field": data["write_field"]}

        def create():
            with session_factory() as session:
                author = models.Author(**author_data)
                session.add(author)
                session.flush()
                session.rollback()

        benchmark(create)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="create-single-author")
    def test_pydantic_sqlalchemy_create(
        self, benchmark, session_factory, create_author_data
    ):
        """Pydantic + SQLAlchemy: Validate then create."""
        author_data = random.choice(create_author_data)

        def create():
            with session_factory() as session:
                # Validate with Pydantic
                validated = schemas.AuthorCreate.model_validate(author_data)
                # Dump to dict and create ORM
                orm_author = models.Author(**validated.model_dump())
                session.add(orm_author)
                session.flush()
                session.rollback()

        benchmark(create)

    @pytest.mark.benchmark(group="create-single-author")
    def test_arcanus_create(
        self, benchmark, arcanus_session_factory, create_author_data
    ):
        """arcanus: Transmuter creation."""
        data = random.choice(create_author_data)
        author_data = {"name": data["name"], "field": data["write_field"]}

        def create():
            with arcanus_session_factory() as session:
                author = Author(**author_data)
                session.add(author)
                session.flush()
                session.rollback()

        benchmark(create)


class TestCreateNestedBook:
    """
    Benchmark book creation with nested author and publisher.

    Simulates: POST /api/books with nested author/publisher data
    Each iteration creates one book with nested objects from random selection.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="create-nested-book")
    def test_sqlalchemy_create_nested(
        self, benchmark, session_factory, create_nested_book_data
    ):
        """Pure SQLAlchemy: Create with nested objects."""
        book_data = random.choice(create_nested_book_data)

        def create():
            with session_factory() as session:
                # Create nested objects manually
                author = models.Author(**book_data["author"])
                publisher = models.Publisher(**book_data["publisher"])
                book = models.Book(
                    title=book_data["title"],
                    year=book_data["year"],
                    author=author,
                    publisher=publisher,
                )
                session.add(book)
                session.flush()
                session.rollback()

        benchmark(create)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="create-nested-book")
    def test_pydantic_sqlalchemy_create_nested(
        self, benchmark, session_factory, create_nested_book_data
    ):
        """Pydantic + SQLAlchemy: Validate then create."""
        book_data = random.choice(create_nested_book_data)

        def create():
            with session_factory() as session:
                # Validate with Pydantic (validates nested structures too)
                validated = schemas.BookCreate.model_validate(book_data)
                # Create ORM objects from validated data
                author = models.Author(**validated.author.model_dump())
                publisher = models.Publisher(**validated.publisher.model_dump())
                book = models.Book(
                    title=validated.title,
                    year=validated.year,
                    author=author,
                    publisher=publisher,
                )
                session.add(book)
                session.flush()
                session.rollback()

        benchmark(create)

    @pytest.mark.benchmark(group="create-nested-book")
    def test_arcanus_create_nested(
        self, benchmark, arcanus_session_factory, create_nested_book_data
    ):
        """arcanus: Transmuter with nested objects."""
        book_data = random.choice(create_nested_book_data)

        def create():
            with arcanus_session_factory() as session:
                book = Book(**book_data)
                session.add(book)
                session.flush()
                session.rollback()

        benchmark(create)


class TestReadSingleAuthor:
    """
    Benchmark reading a single author by ID.

    Simulates: GET /api/authors/{id}
    Each iteration reads one randomly selected author.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="read-single-author")
    def test_sqlalchemy_read(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pure SQLAlchemy: Query by ID."""
        author = random.choice(seeded_authors)
        author_id = author.id

        def read():
            with session_factory() as session:
                result = session.get(models.Author, author_id)
                assert result is not None

        benchmark(read)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="read-single-author")
    def test_pydantic_sqlalchemy_read(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pydantic + SQLAlchemy: Query then validate."""
        author = random.choice(seeded_authors)
        author_id = author.id

        def read():
            with session_factory() as session:
                orm_author = session.get(models.Author, author_id)
                # Convert to Pydantic for API response
                validated = schemas.AuthorFlat.model_validate(orm_author)
                assert validated is not None

        benchmark(read)

    @pytest.mark.benchmark(group="read-single-author")
    def test_arcanus_read(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Query returns transmuter."""
        author = random.choice(seeded_authors)
        author_id = author.id

        def read():
            with arcanus_session_factory() as session:
                result = session.get(Author, author_id)
                assert result is not None

        benchmark(read)


class TestReadManyAuthors:
    """
    Benchmark reading multiple authors with list query.

    Simulates: GET /api/authors?limit=50
    Each iteration fetches 50 authors.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="read-many-authors")
    def test_sqlalchemy_read_many(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pure SQLAlchemy: Query list."""

        def read():
            with session_factory() as session:
                stmt = (
                    select(models.Author)
                    .options(selectinload(models.Author.books))
                    .limit(BATCH_SIZE)
                )
                result = session.scalars(stmt).all()
                assert len(result) == BATCH_SIZE

        benchmark(read)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="read-many-authors")
    def test_pydantic_sqlalchemy_read_many(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pydantic + SQLAlchemy: Query then validate list."""

        def read():
            with session_factory() as session:
                stmt = (
                    select(models.Author)
                    .options(selectinload(models.Author.books))
                    .limit(BATCH_SIZE)
                )
                orm_authors = session.scalars(stmt).all()
                # Convert to Pydantic for API response
                validated = [schemas.AuthorFlat.model_validate(a) for a in orm_authors]
                assert len(validated) == BATCH_SIZE

        benchmark(read)

    @pytest.mark.benchmark(group="read-many-authors")
    def test_arcanus_read_many(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Query returns transmuters."""

        def read():
            with arcanus_session_factory() as session:
                stmt = (
                    select(Author)
                    .options(selectinload(Author["books"]))
                    .limit(BATCH_SIZE)
                )
                result = session.scalars(stmt).all()
                assert len(result) == BATCH_SIZE

        benchmark(read)


class TestReadNestedBook:
    """
    Benchmark reading books with eager-loaded relationships.

    Simulates: GET /api/books/{id} with author and publisher included
    Each iteration reads one randomly selected book with relationships.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="read-nested-book")
    def test_sqlalchemy_read_nested(
        self, benchmark, session_factory, seeded_books: list[models.Book]
    ):
        """Pure SQLAlchemy: Query with joinedload."""
        book = random.choice(seeded_books)
        book_id = book.id

        def read():
            with session_factory() as session:
                stmt = (
                    select(models.Book)
                    .where(models.Book.id == book_id)
                    .options(
                        joinedload(models.Book.author),
                        joinedload(models.Book.publisher),
                    )
                )
                result = session.scalars(stmt).first()
                assert result is not None
                assert result.author is not None
                assert result.publisher is not None

        benchmark(read)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="read-nested-book")
    def test_pydantic_sqlalchemy_read_nested(
        self, benchmark, session_factory, seeded_books: list[models.Book]
    ):
        """Pydantic + SQLAlchemy: Query then validate."""
        book = random.choice(seeded_books)
        book_id = book.id

        def read():
            with session_factory() as session:
                stmt = (
                    select(models.Book)
                    .where(models.Book.id == book_id)
                    .options(
                        joinedload(models.Book.author),
                        joinedload(models.Book.publisher),
                    )
                )
                orm_book = session.scalars(stmt).first()
                # Convert to Pydantic for API response
                validated = schemas.BookFlat.model_validate(orm_book)
                assert validated is not None
                assert validated.author is not None

        benchmark(read)

    @pytest.mark.benchmark(group="read-nested-book")
    def test_arcanus_read_nested(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_books: list[models.Book],
    ):
        """arcanus: Query with eager load."""
        book = random.choice(seeded_books)
        book_id = book.id

        def read():
            with arcanus_session_factory() as session:
                stmt = (
                    select(Book)
                    .where(Book["id"] == book_id)
                    .options(
                        joinedload(Book["author"]),
                        joinedload(Book["publisher"]),
                    )
                )
                result = session.scalars(stmt).first()
                assert result is not None
                assert result.author.value is not None

        benchmark(read)


class TestUpdateSingleAuthor:
    """
    Benchmark updating a single author.

    Simulates: PUT/PATCH /api/authors/{id}
    Each iteration updates one randomly selected author.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="update-single-author")
    def test_sqlalchemy_update(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pure SQLAlchemy: Direct ORM update."""
        author = random.choice(seeded_authors)
        author_id = author.id

        def update():
            with session_factory() as session:
                orm_author = session.get(models.Author, author_id)
                orm_author.name = "Updated Name"
                orm_author.field = "Physics"
                session.flush()
                session.rollback()

        benchmark(update)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="update-single-author")
    def test_pydantic_sqlalchemy_update(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pydantic + SQLAlchemy: Validate then update."""
        author = random.choice(seeded_authors)
        author_id = author.id
        update_data = {"name": "Updated Name", "write_field": "Physics"}

        def update():
            with session_factory() as session:
                # Validate update data
                validated = schemas.AuthorCreate.model_validate(update_data)
                # Fetch and update ORM object
                orm_author = session.get(models.Author, author_id)
                for k, v in validated.model_dump(
                    by_alias=True, exclude_unset=True
                ).items():
                    setattr(orm_author, k, v)
                session.flush()
                session.rollback()

        benchmark(update)

    @pytest.mark.benchmark(group="update-single-author")
    def test_arcanus_update(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Transmuter update."""
        author = random.choice(seeded_authors)
        author_id = author.id
        update_data = {"name": "Updated Name", "write_field": "Physics"}
        AuthorUpdate = Author.Update

        def update():
            with arcanus_session_factory() as session:
                updated = AuthorUpdate(**update_data)
                transmuter = session.get(Author, author_id)
                transmuter.absorb(updated)
                session.flush()
                session.rollback()

        benchmark(update)


class TestUpdateManyAuthors:
    """
    Benchmark bulk update of multiple authors.

    Simulates: Batch update endpoint
    Each iteration updates 50 authors.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="update-many-authors")
    def test_sqlalchemy_update_many(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pure SQLAlchemy: Bulk update."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def update():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                for author in authors:
                    author.name = f"Bulk Updated {author.id}"
                    author.field = "Physics"
                session.flush()
                session.rollback()

        benchmark(update)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="update-many-authors")
    def test_pydantic_sqlalchemy_update_many(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pydantic + SQLAlchemy: Validate then update."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def update():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                for author in authors:
                    update_data = {
                        "name": f"Bulk Updated {author.id}",
                        "write_field": "Physics",
                    }
                    validated = schemas.AuthorCreate.model_validate(update_data)
                    for k, v in validated.model_dump(
                        by_alias=True, exclude_unset=True
                    ).items():
                        setattr(author, k, v)
                session.flush()
                session.rollback()

        benchmark(update)

    @pytest.mark.benchmark(group="update-many-authors")
    def test_arcanus_update_many(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Bulk update."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]
        AuthorUpdate = Author.Update

        def update():
            with arcanus_session_factory() as session:
                stmt = select(Author).where(Author["id"].in_(author_ids))
                authors = session.scalars(stmt).all()
                for author in authors:
                    updated = AuthorUpdate(
                        name=f"Bulk Updated {author.id}",
                        write_field="Physics",
                    )
                    author.absorb(updated)
                session.flush()
                session.rollback()

        benchmark(update)


class TestSerializeToDict:
    """
    Benchmark serializing ORM/transmuter to dict for API response.

    Simulates: Converting query result to JSON response
    Each iteration serializes 50 authors.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="serialize-dict")
    def test_sqlalchemy_serialize_dict(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pure SQLAlchemy: Manual dict."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def serialize():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                return [{"id": a.id, "name": a.name, "field": a.field} for a in authors]

        result = benchmark(serialize)
        assert len(result) == BATCH_SIZE

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="serialize-dict")
    def test_pydantic_sqlalchemy_serialize_dict(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pydantic + SQLAlchemy: Validate then dump."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def serialize():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                validated = [schemas.AuthorFlat.model_validate(a) for a in authors]
                return [v.model_dump() for v in validated]

        result = benchmark(serialize)
        assert len(result) == BATCH_SIZE

    @pytest.mark.benchmark(group="serialize-dict")
    def test_arcanus_serialize_dict(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: model_dump."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def serialize():
            with arcanus_session_factory() as session:
                stmt = select(Author).where(Author["id"].in_(author_ids))
                authors = session.scalars(stmt).all()
                return [a.model_dump(exclude={"books", "test_id"}) for a in authors]

        result = benchmark(serialize)
        assert len(result) == BATCH_SIZE


class TestSerializeToJson:
    """
    Benchmark serializing ORM/transmuter to JSON string for API response.

    Simulates: Direct JSON response generation
    Each iteration serializes 50 authors to JSON.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="serialize-json")
    def test_sqlalchemy_serialize_json(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pure SQLAlchemy: Manual JSON conversion via json module."""
        import json

        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def serialize():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                data = [{"id": a.id, "name": a.name, "field": a.field} for a in authors]
                return json.dumps(data)

        result = benchmark(serialize)
        assert len(result) > 0

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="serialize-json")
    def test_pydantic_sqlalchemy_serialize_json(
        self,
        benchmark,
        session_factory,
        seeded_authors: list[models.Author],
    ):
        """Pydantic + SQLAlchemy: Validate then model_dump_json."""
        from pydantic import TypeAdapter

        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]
        adapter = TypeAdapter(list[schemas.AuthorFlat])

        def serialize():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                validated = adapter.validate_python(authors)
                return adapter.dump_json(validated)

        result = benchmark(serialize)
        assert len(result) > 0

    @pytest.mark.benchmark(group="serialize-json")
    def test_arcanus_serialize_json(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Transmuter model_dump_json."""
        from pydantic import TypeAdapter

        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]
        adapter = TypeAdapter(list[Author])

        def serialize():
            with arcanus_session_factory() as session:
                stmt = select(Author).where(Author["id"].in_(author_ids))
                authors = session.scalars(stmt).all()
                return adapter.dump_json(
                    authors, exclude={"__all__": {"books", "test_id"}}
                )

        result = benchmark(serialize)
        assert len(result) > 0


class TestScalarsOnlySerializeToDict:
    """
    Benchmark scalar-only model_dump() with SQLAlchemy materia.

    Uses ORM objects converted to transmuters, then dumped to dict.
    Tests the serialization path without relationship overhead.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="serialize-scalars-only-dict")
    def test_pydantic_sqlalchemy_scalars_dump_dict(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pydantic + SQLAlchemy: Load ORM, validate to flat model, dump to dict."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def dump_all():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                validated = [schemas.AuthorFlat.model_validate(a) for a in authors]
                return [v.model_dump() for v in validated]

        result = benchmark(dump_all)
        assert len(result) == BATCH_SIZE
        assert "name" in result[0]

    @pytest.mark.benchmark(group="serialize-scalars-only-dict")
    def test_arcanus_scalars_dump_dict(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Load transmuter, dump to dict (excluding relationships)."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def dump_all():
            with arcanus_session_factory() as session:
                stmt = select(Author).where(Author["id"].in_(author_ids))
                authors = session.scalars(stmt).all()
                return [a.model_dump(exclude={"books"}) for a in authors]

        result = benchmark(dump_all)
        assert len(result) == BATCH_SIZE
        assert "name" in result[0]


class TestScalarsOnlySerializeToJson:
    """
    Benchmark scalar-only model_dump_json() with SQLAlchemy materia.

    Uses ORM objects converted to transmuters, then dumped to JSON.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="serialize-scalars-only-json")
    def test_pydantic_sqlalchemy_scalars_dump_json(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pydantic + SQLAlchemy: Load ORM, validate to flat model, dump to JSON."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def dump_all():
            with session_factory() as session:
                stmt = select(models.Author).where(models.Author.id.in_(author_ids))
                authors = session.scalars(stmt).all()
                validated = [schemas.AuthorFlat.model_validate(a) for a in authors]
                return [v.model_dump_json() for v in validated]

        result = benchmark(dump_all)
        assert len(result) == BATCH_SIZE

    @pytest.mark.benchmark(group="serialize-scalars-only-json")
    def test_arcanus_scalars_dump_json(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Load transmuter, dump to JSON (excluding relationships)."""
        author_ids = [a.id for a in seeded_authors[:BATCH_SIZE]]

        def dump_all():
            with arcanus_session_factory() as session:
                stmt = select(Author).where(Author["id"].in_(author_ids))
                authors = session.scalars(stmt).all()
                return [a.model_dump_json(exclude={"books"}) for a in authors]

        result = benchmark(dump_all)
        assert len(result) == BATCH_SIZE


class TestRoundtripAuthor:
    """
    Benchmark full roundtrip: Load -> Modify -> Save.

    Simulates: GET /api/authors/{id}, modify, PUT /api/authors/{id}
    Each iteration performs one roundtrip on a randomly selected author.
    """

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="roundtrip-author")
    def test_sqlalchemy_roundtrip(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pure SQLAlchemy: Load, modify, flush, rollback."""
        author = random.choice(seeded_authors)
        author_id = author.id

        def roundtrip():
            with session_factory() as session:
                orm_author = session.get(models.Author, author_id)
                orm_author.name = f"Roundtrip Updated {author_id}"
                orm_author.field = "Physics"
                session.flush()
                session.rollback()

        benchmark(roundtrip)

    @pytest.mark.baseline
    @pytest.mark.benchmark(group="roundtrip-author")
    def test_pydantic_sqlalchemy_roundtrip(
        self, benchmark, session_factory, seeded_authors: list[models.Author]
    ):
        """Pydantic + SQLAlchemy: Load, validate input, apply, flush."""
        author = random.choice(seeded_authors)
        author_id = author.id

        def roundtrip():
            with session_factory() as session:
                orm_author = session.get(models.Author, author_id)
                update_data = {
                    "name": f"Roundtrip Updated {author_id}",
                    "write_field": "Physics",
                }
                validated = schemas.AuthorCreate.model_validate(update_data)
                orm_author.name = validated.name
                orm_author.field = validated.field
                session.flush()
                session.rollback()

        benchmark(roundtrip)

    @pytest.mark.benchmark(group="roundtrip-author")
    def test_arcanus_roundtrip(
        self,
        benchmark,
        arcanus_session_factory,
        seeded_authors: list[models.Author],
    ):
        """arcanus: Load transmuter, modify, flush."""
        author = random.choice(seeded_authors)
        author_id = author.id

        def roundtrip():
            with arcanus_session_factory() as session:
                transmuter = session.get(Author, author_id)
                transmuter.name = f"Roundtrip Updated {author_id}"
                transmuter.field = "Physics"
                session.flush()
                session.rollback()

        benchmark(roundtrip)



from __future__ import annotations

# SQLAlchemy Materia test fixtures
# This module contains shared fixtures for testing the SQLAlchemy materia implementation.
# Includes database engine setup, table creation, and data fixtures for relationships testing.
# These fixtures are shared between sync and async test modules.
import pytest
import pytest_asyncio
from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from arcanus.association import Relation, RelationCollection
from arcanus.materia.sqlalchemy import AsyncSession
from tests.models import Base
from tests.transmuters import (
    Author,
    Book,
    BookDetail,
    Category,
    Publisher,
    Review,
    Translator,
    sqlalchemy_materia,
)

# Use SQLite in-memory for tests - avoids external database dependency
# Using shared cache mode to allow multiple connections to same in-memory database
DB_URL = "sqlite:///:memory:?cache=shared"
# DB_URL = "postgresql+psycopg2://postgres:postgres@localhost:5432/arcanus"  # For local testing with Postgres
ASYNC_DB_URL = "sqlite+aiosqlite:///:memory:?cache=shared"
# ASYNC_DB_URL = "postgresql+asyncpg://postgres:postgres@localhost:5432/arcanus"  # For local testing with Postgres


@pytest.fixture(scope="module", autouse=True)
def materia():
    with sqlalchemy_materia:
        yield sqlalchemy_materia


@pytest_asyncio.fixture(scope="session")
async def async_engine():
    engine = create_async_engine(
        ASYNC_DB_URL,
        # echo=True,
        future=True,
    )
    # Create tables for async tests
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    try:
        yield engine
    finally:
        await engine.dispose()


@pytest.fixture(scope="session")
def engine():
    sync_engine = create_engine(
        DB_URL,
        # echo=True,
        future=True,
    )

    try:
        yield sync_engine
    finally:
        sync_engine.dispose()


@pytest.fixture(scope="session", autouse=True)
def setup_database(engine):
    with engine.begin() as conn:
        Base.metadata.drop_all(conn)
        Base.metadata.create_all(conn)


@pytest_asyncio.fixture(scope="session")
async def publisher(async_engine: AsyncEngine) -> Publisher:
    """Persist and return a Publisher without books."""
    async with AsyncSession(async_engine) as session:
        pub = Publisher(name="Cambridge University Press", country="United Kingdom")
        session.add(pub)
        await session.flush()
        await session.commit()
        await session.refresh(pub)
        return pub


@pytest_asyncio.fixture(scope="session")
async def author(async_engine: AsyncEngine) -> Author:
    """Persist and return an Author without books."""
    async with AsyncSession(async_engine) as session:
        auth = Author(name="Stephen Hawking", field="Physics")
        session.add(auth)
        await session.flush()
        await session.commit()
        await session.refresh(auth)
        return auth


@pytest_asyncio.fixture(scope="session")
async def category(async_engine: AsyncEngine) -> Category:
    """Persist and return a Category."""
    async with AsyncSession(async_engine) as session:
        cat = Category(
            name="Cosmology",
            description="Study of the origin and evolution of the universe",
        )
        session.add(cat)
        await session.flush()
        await session.commit()
        await session.refresh(cat)
        return cat


@pytest_asyncio.fixture(scope="session")
async def book_with_relations(async_engine: AsyncEngine) -> Book:
    """Persist and return a Book with all relationships (1-1, M-1, M-M)."""
    async with AsyncSession(async_engine) as session:
        # Create Book with M-1 relationships
        book = Book(
            title="A Brief History of Time",
            year=1988,
            author=Relation(Author(name="Stephen Hawking", field="Physics")),
            publisher=Relation(Publisher(name="Bantam Books", country="United States")),
            categories=RelationCollection(
                [
                    Category(
                        name="Theoretical Physics",
                        description="Study of physical theories",
                    ),
                    Category(
                        name="Popular Science",
                        description="Science for general audience",
                    ),
                ]
            ),
            detail=Relation(
                BookDetail(
                    isbn="978-0553380163",
                    pages=212,
                    abstract="A landmark volume in science writing by one of the great minds of our time.",
                ),
            ),
        )
        session.add(book)
        await session.flush()
        await session.commit()
        await session.refresh(book)
        return book


@pytest_asyncio.fixture(scope="session")
async def author_with_books(async_engine: AsyncEngine) -> Author:
    """Persist and return an Author with multiple Books (one-to-many)."""
    async with AsyncSession(async_engine) as session:
        pub = Publisher(name="W. W. Norton", country="United States")
        session.add(pub)

        auth = Author(name="Richard Feynman", field="Quantum Physics")
        session.add(auth)
        await session.flush()

        book1 = Book(
            title="The Feynman Lectures on Physics",
            year=1964,
            author=Relation(auth),
            publisher=Relation(pub),
        )
        book2 = Book(
            title="QED: The Strange Theory of Light and Matter",
            year=1985,
            author=Relation(auth),
            publisher=Relation(pub),
        )
        session.add(book1)
        session.add(book2)
        await session.flush()
        await session.commit()
        await session.refresh(auth)
        return auth


@pytest_asyncio.fixture()
async def book_with_translator(async_engine: AsyncEngine) -> Book:
    """Persist and return a Book with an optional Translator (1-1 optional relationship)."""
    async with AsyncSession(async_engine) as session:
        pub = Publisher(name="Springer", country="Germany")
        session.add(pub)

        auth = Author(name="Albert Einstein", field="Physics")
        session.add(auth)

        # Create a Translator (optional 1-1)
        translator = Translator(name="Robert W. Lawson", language="English")
        session.add(translator)
        await session.flush()

        # Create a translated book
        book = Book(
            title="Relativity: The Special and General Theory",
            year=1920,
            author=Relation(auth),
            publisher=Relation(pub),
            translator=Relation(translator),  # Optional 1-1 relationship
        )
        session.add(book)
        await session.flush()
        await session.commit()
        await session.refresh(book)
        return book


@pytest_asyncio.fixture()
async def book_with_reviews(async_engine: AsyncEngine) -> Book:
    """Persist and return a Book with optional Reviews (1-M optional relationship)."""
    async with AsyncSession(async_engine) as session:
        pub = Publisher(name="Vintage Books", country="United States")
        session.add(pub)

        auth = Author(name="James Watson", field="Biology")
        session.add(auth)
        await session.flush()

        book = Book(
            title="The Double Helix",
            year=1968,
            author=Relation(auth),
            publisher=Relation(pub),
        )
        session.add(book)
        await session.flush()

        # Add optional reviews (1-M)
        review1 = Review(
            reviewer_name="Nature Magazine",
            rating=5,
            comment="A groundbreaking account of DNA structure discovery.",
            book=Relation(book),
        )
        review2 = Review(
            reviewer_name="Science Journal",
            rating=4,
            comment="An insightful personal narrative of scientific discovery.",
            book=Relation(book),
        )
        session.add(review1)
        session.add(review2)
        await session.flush()
        await session.commit()
        await session.refresh(book)
        return book



from functools import lru_cache
from typing import TypeVar

from pydantic import TypeAdapter

T = TypeVar("T")


@lru_cache(maxsize=100)
def get_cached_adapter(type: type[T]) -> TypeAdapter[T]:
    """Cached TypeAdapter factory to avoid recreating adapters for the same type."""
    return TypeAdapter(type)



from __future__ import annotations

import random
from typing import Any

import pytest

from tests import schemas
from tests import transmuters as transmuters_module


def pytest_configure(config):
    """Register custom markers."""
    config.addinivalue_line(
        "markers",
        "baseline: mark test as baseline/control group (excluded from CodSpeed regression checks)",
    )


SEED = 42
random.seed(SEED)


@pytest.fixture(scope="session")
def simple_author_data() -> list[dict[str, Any]]:
    """Generate simple author data without relationships (once per session)."""
    random.seed(SEED)
    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    names = [
        "Isaac Asimov",
        "Arthur C. Clarke",
        "Ursula K. Le Guin",
        "Robert A. Heinlein",
        "Frank Herbert",
        "William Gibson",
        "Octavia E. Butler",
        "Ray Bradbury",
        "Larry Niven",
        "Neal Stephenson",
        "Philip K. Dick",
        "Cixin Liu",
    ]
    return [
        {
            "id": i,
            "name": f"{random.choice(names)} {random.randint(100, 999)}",
            "field": random.choice(fields),
        }
        for i in range(100)
    ]


@pytest.fixture(scope="session")
def nested_book_data() -> list[dict[str, Any]]:
    """Generate book data with one level of nested relationships (once per session)."""
    random.seed(SEED)
    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    scifi_authors = [
        "Isaac Asimov",
        "Arthur C. Clarke",
        "Ursula K. Le Guin",
        "Robert A. Heinlein",
        "Frank Herbert",
        "William Gibson",
        "Octavia E. Butler",
        "Ray Bradbury",
        "Larry Niven",
        "Neal Stephenson",
        "Philip K. Dick",
        "Cixin Liu",
    ]
    countries = ["USA", "UK", "Germany", "France", "Japan", "Canada"]
    adjectives = [
        "Galactic",
        "Neon",
        "Stellar",
        "Interstellar",
        "Synthetic",
        "Cybernetic",
    ]
    nouns = ["Odyssey", "Chronicles", "Protocol", "Singularity", "Nexus", "Expedition"]

    books = []
    for i in range(50):
        book = {
            "id": i,
            "title": f"The {random.choice(adjectives)} {random.choice(nouns)} {random.randint(1, 99)}",
            "year": random.randint(1990, 2024),
            "author": {
                "id": i,
                "name": f"{random.choice(scifi_authors)} {random.randint(100, 999)}",
                "field": random.choice(fields),
            },
            "publisher": {
                "id": i % 5,
                "name": f"Publisher {random.choice(['House', 'Press', 'Books'])} {random.randint(1, 50)}",
                "country": random.choice(countries),
            },
        }
        books.append(book)

    return books


@pytest.fixture(scope="session")
def deep_nested_data() -> list[dict[str, Any]]:
    """
    Generate author data with deeply nested book relationships (once per session).

    Creates author -> books -> author structure, but breaks circular reference
    by having book.author contain only scalar fields (no books array) to avoid
    infinite recursion during Pydantic validation.

    Only includes author, publisher to avoid complex nested requirements
    (e.g., BookDetail.book is required and would create recursion issues).
    """
    random.seed(SEED)
    author_count = 10
    books_per_author = 10
    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    scifi_authors = [
        "Isaac Asimov",
        "Arthur C. Clarke",
        "Ursula K. Le Guin",
        "Robert A. Heinlein",
        "Frank Herbert",
        "William Gibson",
        "Octavia E. Butler",
        "Ray Bradbury",
        "Larry Niven",
        "Neal Stephenson",
        "Philip K. Dick",
        "Cixin Liu",
    ]
    countries = ["USA", "UK", "Germany", "France", "Japan"]
    adjectives = [
        "Galactic",
        "Neon",
        "Stellar",
        "Interstellar",
        "Synthetic",
        "Cybernetic",
    ]
    nouns = ["Odyssey", "Chronicles", "Protocol", "Singularity", "Nexus", "Expedition"]

    authors = []
    for i in range(author_count):
        num_books = books_per_author + random.randint(-2, 2)
        num_books = max(1, num_books)

        author_id = i
        author_name = f"{random.choice(scifi_authors)} {random.randint(100, 999)} {random.choice(['Sr.', 'Jr.', 'PhD', ''])}"
        author_field = random.choice(fields)

        books = []
        for j in range(num_books):
            book_dict = {
                "id": i * books_per_author + j,
                "title": f"The {random.choice(adjectives)} {random.choice(nouns)}",
                "year": random.randint(1995, 2024),
                "author": {
                    "id": author_id,
                    "name": author_name,
                    "field": author_field,
                },
                "publisher": {
                    "id": j % 5,
                    "name": f"Publisher {random.randint(1, 50)}",
                    "country": random.choice(countries),
                },
            }
            books.append(book_dict)

        author_dict = {
            "id": author_id,
            "name": author_name,
            "field": author_field,
            "books": books,
        }
        authors.append(author_dict)
    return authors


@pytest.fixture(scope="session")
def circular_data() -> list[dict[str, Any]]:
    """
    Generate company data with true circular references (Company -> Dept -> Employee -> back).

    Creates proper circular references between entities (for transmuter-only tests):
    - Company references its CEO (an Employee)
    - Departments reference their Company
    - Employees reference their Department and Company

    NOTE: Pydantic cannot handle these circular references due to infinite recursion.
    This fixture is intended for NoOpMateria transmuter tests only.
    """
    random.seed(SEED)
    company_count = 5
    industries = ["Tech", "Finance", "Healthcare", "Retail", "Manufacturing"]
    titles = ["Engineer", "Manager", "Analyst", "Director", "VP"]
    dept_names = ["Engineering", "Sales", "Marketing", "HR", "Finance", "R&D"]

    companies = []
    for c in range(company_count):
        num_depts = random.randint(2, 4)

        company_dict: dict[str, Any] = {
            "id": c,
            "name": f"Company {random.randint(100, 999)}",
            "industry": random.choice(industries),
        }

        departments = []
        all_employees = []

        for d in range(num_depts):
            num_employees = random.randint(3, 6)

            dept_dict: dict[str, Any] = {
                "id": c * 10 + d,
                "name": f"{random.choice(dept_names)} {random.randint(1, 99)}",
                "budget": random.randint(100000, 5000000),
                "company": company_dict,
            }

            employees = []
            for e in range(num_employees):
                employee_dict = {
                    "id": c * 100 + d * 10 + e,
                    "name": f"Employee {random.randint(100, 999)}",
                    "title": random.choice(titles),
                    "salary": random.randint(50000, 200000),
                    "department": dept_dict,
                    "company": company_dict,
                }
                employees.append(employee_dict)

            dept_dict["employees"] = employees
            all_employees.extend(employees)
            departments.append(dept_dict)

        company_dict["departments"] = departments
        company_dict["employees"] = all_employees
        company_dict["ceo"] = random.choice(all_employees) if all_employees else None

        companies.append(company_dict)
    return companies


@pytest.fixture(scope="session")
def simple_author_models(
    simple_author_data: list[dict[str, Any]],
) -> list[schemas.Author]:
    """Pre-validated Pydantic author models."""
    return [schemas.Author.model_validate(d) for d in simple_author_data]


@pytest.fixture(scope="session")
def simple_author_transmuters(
    simple_author_data: list[dict[str, Any]],
) -> list[transmuters_module.Author]:
    """Pre-validated NoOp author transmuters."""
    return [transmuters_module.Author.model_validate(d) for d in simple_author_data]


@pytest.fixture(scope="session")
def simple_author_flat_models(
    simple_author_data: list[dict[str, Any]],
) -> list[schemas.AuthorFlat]:
    """Pre-validated flat Pydantic author models (no relationships)."""
    return [schemas.AuthorFlat.model_validate(d) for d in simple_author_data]


@pytest.fixture(scope="session")
def simple_author_flat_transmuters(
    simple_author_data: list[dict[str, Any]],
) -> list[transmuters_module.AuthorFlat]:
    """Pre-validated flat NoOp author transmuters (no relationships)."""
    return [transmuters_module.AuthorFlat.model_validate(d) for d in simple_author_data]


class MockAuthor:
    """Mock ORM-like author object."""

    def __init__(self, id: int, name: str, field: str):
        self.id = id
        self.name = name
        self.field = field


class MockPublisher:
    """Mock ORM-like publisher object."""

    def __init__(self, id: int, name: str, country: str):
        self.id = id
        self.name = name
        self.country = country


class MockBook:
    """Mock ORM-like book object with nested relationships."""

    def __init__(
        self,
        id: int,
        title: str,
        year: int,
        author: MockAuthor,
        publisher: MockPublisher,
    ):
        self.id = id
        self.title = title
        self.year = year
        self.author = author
        self.publisher = publisher


@pytest.fixture(scope="session")
def mock_author_objects() -> list[MockAuthor]:
    """Mock ORM-like author objects."""
    random.seed(SEED)
    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    names = [
        "Isaac Asimov",
        "Arthur C. Clarke",
        "Ursula K. Le Guin",
        "Robert A. Heinlein",
        "Frank Herbert",
        "William Gibson",
    ]
    return [
        MockAuthor(
            i,
            f"{random.choice(names)} {random.randint(1, 999)}",
            random.choice(fields),
        )
        for i in range(100)
    ]


@pytest.fixture(scope="session")
def mock_nested_book_objects() -> list[MockBook]:
    """Mock ORM-like book objects with nested relationships."""
    random.seed(SEED)
    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    countries = ["USA", "UK", "Germany", "France", "Japan"]
    adjectives = ["Galactic", "Neon", "Stellar", "Interstellar", "Synthetic"]

    books = []
    for i in range(100):
        author = MockAuthor(
            i % 10, f"Author {random.randint(1, 999)}", random.choice(fields)
        )
        publisher = MockPublisher(
            i % 5, f"Publisher {random.randint(1, 50)}", random.choice(countries)
        )
        book = MockBook(
            i,
            f"The {random.choice(adjectives)} Book {random.randint(1, 99)}",
            random.randint(1990, 2024),
            author,
            publisher,
        )
        books.append(book)
    return books



from __future__ import annotations

import random
from typing import Any, Generator

import pytest
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import sessionmaker

from arcanus.materia.sqlalchemy import Session as arcanusSession
from arcanus.materia.sqlalchemy.base import SqlalchemyMateria
from tests.models import Author as AuthorModel
from tests.models import Base
from tests.models import Book as BookModel
from tests.models import Publisher as PublisherModel
from tests.transmuters import sqlalchemy_materia

SEED = 42
# Use SQLite in-memory for benchmarks - avoids network I/O for instrumentation mode
DB_URL = "sqlite:///:memory:"


@pytest.fixture(scope="session")
def engine() -> Generator[Engine, None, None]:
    """Create a database engine for benchmark tests."""
    sync_engine = create_engine(
        DB_URL,
        echo=False,
        future=True,
    )

    try:
        yield sync_engine
    finally:
        sync_engine.dispose()


@pytest.fixture(scope="session", autouse=True)
def setup_database(engine: Engine) -> None:
    """Set up the database schema."""
    with engine.begin() as conn:
        Base.metadata.drop_all(conn)
        Base.metadata.create_all(conn)


@pytest.fixture(scope="session")
def session_factory(engine: Engine) -> sessionmaker:
    """Create a session factory."""
    return sessionmaker(bind=engine, expire_on_commit=False)


@pytest.fixture(scope="session")
def arcanus_session_factory(engine: Engine) -> sessionmaker:
    """Create an arcanus session factory."""
    return sessionmaker(bind=engine, expire_on_commit=False, class_=arcanusSession)


@pytest.fixture(scope="module", autouse=True)
def materia() -> Generator[SqlalchemyMateria, None, None]:
    """Activate sqlalchemy_materia for tests that need it."""
    with sqlalchemy_materia:
        yield sqlalchemy_materia


@pytest.fixture(scope="session")
def seeded_authors(
    session_factory: sessionmaker,
) -> Generator[list[AuthorModel], None, None]:
    """
    Pre-seed the database with authors for read/update/delete benchmarks.

    Creates 100 authors with reproducible data for consistent benchmarks.
    This fixture is session-scoped - data persists across all tests.
    """
    random.seed(SEED)
    session = session_factory()

    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    names = [
        "Isaac Asimov",
        "Arthur C. Clarke",
        "Ursula K. Le Guin",
        "Robert A. Heinlein",
        "Frank Herbert",
        "William Gibson",
    ]

    authors = []
    for _ in range(100):
        author = AuthorModel(
            name=f"{random.choice(names)} {random.randint(100, 999)}",
            field=random.choice(fields),
        )
        session.add(author)
        authors.append(author)

    session.commit()

    for author in authors:
        session.refresh(author)

    yield authors

    session.close()


@pytest.fixture(scope="session")
def seeded_books(
    session_factory: sessionmaker,
    seeded_authors: list[AuthorModel],
) -> Generator[list[BookModel], None, None]:
    """
    Pre-seed the database with books (with relationships) for read benchmarks.

    Creates 100 books linked to authors and publishers.
    """
    random.seed(SEED + 1)
    session = session_factory()

    countries = ["USA", "UK", "Germany", "France", "Japan", "Canada"]
    adjectives = [
        "Galactic",
        "Neon",
        "Stellar",
        "Interstellar",
        "Synthetic",
        "Cybernetic",
    ]
    nouns = ["Odyssey", "Chronicles", "Protocol", "Singularity", "Nexus", "Expedition"]

    publishers = []
    for i in range(10):
        publisher = PublisherModel(
            name=f"Publisher {random.choice(['House', 'Press', 'Books'])} {i}",
            country=random.choice(countries),
        )
        session.add(publisher)
        publishers.append(publisher)

    session.flush()

    books = []
    for i in range(100):
        author = seeded_authors[i % len(seeded_authors)]
        publisher = publishers[i % len(publishers)]

        book = BookModel(
            title=f"The {random.choice(adjectives)} {random.choice(nouns)} {random.randint(1, 99)}",
            year=random.randint(1990, 2024),
            author_id=author.id,
            publisher_id=publisher.id,
        )
        session.add(book)
        books.append(book)

    session.commit()

    for book in books:
        session.refresh(book)

    yield books

    session.close()


@pytest.fixture(scope="session")
def create_author_data() -> list[dict[str, Any]]:
    """Generate author data for create benchmarks."""
    random.seed(SEED + 3)

    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    names = [
        "Isaac Asimov",
        "Arthur C. Clarke",
        "Ursula K. Le Guin",
        "Robert A. Heinlein",
        "Frank Herbert",
        "William Gibson",
    ]

    return [
        {
            "name": f"{random.choice(names)} {random.randint(100, 999)}",
            "write_field": random.choice(fields),
        }
        for _ in range(50)
    ]


@pytest.fixture(scope="session")
def create_nested_book_data() -> list[dict[str, Any]]:
    """Generate book data with nested author and publisher for create benchmarks."""
    random.seed(SEED + 4)

    fields = [
        "Astrophysics",
        "Robotics",
        "Cybernetics",
        "Xenobiology",
        "Quantum Physics",
        "Science Fiction",
    ]
    countries = ["USA", "UK", "Germany", "France", "Japan", "Canada"]
    adjectives = [
        "Galactic",
        "Neon",
        "Stellar",
        "Interstellar",
        "Synthetic",
        "Cybernetic",
    ]
    nouns = ["Odyssey", "Chronicles", "Protocol", "Singularity", "Nexus", "Expedition"]
    names = [
        "Isaac Asimov",
        "Arthur C. Clarke",
        "Ursula K. Le Guin",
        "Robert A. Heinlein",
        "Frank Herbert",
        "William Gibson",
    ]

    return [
        {
            "title": f"The {random.choice(adjectives)} {random.choice(nouns)} {random.randint(1, 99)}",
            "year": random.randint(1990, 2024),
            "author": {
                "name": f"{random.choice(names)} {random.randint(100, 999)}",
                "field": random.choice(fields),
            },
            "publisher": {
                "name": f"Publisher {random.choice(['House', 'Press', 'Books'])} {random.randint(1, 50)}",
                "country": random.choice(countries),
            },
        }
        for _ in range(50)
    ]
